#!/usr/bin/env python
#
# Convert funpack/scripts/demo/funpack_demonstration.md to
# funpack/scripts/demo/funpack_demonstration.ipynb, then execute it to
# generate funpack/tests/funpack_demonstration_with_outputs.ipynb,
# and doc/demo.rst
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import subprocess as sp
import os.path    as op
import base64     as b64
import               os
import               sys
import               json
import               shlex

import bash_kernel.install as bash_kernel_install

import funpack.scripts.demo  as     demo
import funpack.tests         as     tests
from   funpack.tests         import tempdir


METADATA = {
    'kernelspec' : {
        'display_name' : 'Bash',
        'language'     : 'bash',
        'name'         : 'bash'
    },
    'language_info' : {
        'codemirror_mode' : 'shell',
        'file_extension'  : '.sh',
        'mimetype'        : 'text/x-sh',
        'name'            : 'bash'
    }
}


def create_funpack_executable():
    with open('funpack', 'wt') as f:
        f.write('#!/usr/bin/env bash\n')
        f.write(f'{sys.executable} -m funpack "$@"\n')
    os.chmod('funpack', 0o755)


def encode(imagefile):
    with open(imagefile, 'rb') as f:
        data = f.read()
    return b64.b64encode(data).decode('ascii')


def patch_image(notebook, imagefile):
    filename = op.basename(imagefile)
    found    = False
    for cell in notebook['cells']:
        for lineidx, line in enumerate(cell['source']):
            if f'({filename})' in line:
                found = True
                break
        if found:
            break
    else:
        raise RuntimeError(f'Cannot find cell which refers to {filename}!')

    line = line.replace(f'({filename})', f'(attachment:{filename})')
    cell['source'][lineidx] = line

    cell['attachments'] = {filename : {'image/png' : encode(imagefile)}}


def markdown_to_notebook(demodir, mdfile, nbfile):
    sp.run(shlex.split(f'notedown "{mdfile}" -o "{nbfile}"'), check=True)

    with open(nbfile, 'rt') as f:
        notebook = json.loads(f.read())

    notebook['metadata'] = METADATA

    images = ['win.png', 'coding.png']
    for image in images:
        image = op.join(demodir, image)
        patch_image(notebook, image)

    with open(nbfile, 'wt') as f:
        f.write(json.dumps(notebook, indent=2))


def execute_notebook(nbfile, outfile):
    bash_kernel_install.main([])
    with tempdir():
        create_funpack_executable()
        cmd = f'jupyter-nbconvert {nbfile} --execute ' \
               '--ExecutePreprocessor.kernel_name=bash ' \
              f'--to=notebook --output={outfile}'
        env         = os.environ.copy()
        env['PATH'] = op.pathsep.join((os.getcwd(), env.get('PATH', '')))
        sp.run(shlex.split(cmd), check=True, env=env)


def truncate_long_output_cells(infile, outfile, maxlines=50):
    with open(infile, 'rt') as f:
        notebook = json.loads(f.read())

    for cell in notebook['cells']:
        if cell['cell_type'] != 'code':
            continue
        if 'outputs' not in cell:
            continue
        for output in cell['outputs']:
            if len(output['text']) > maxlines:
                ntrunc = len(output['text']) - maxlines
                print('truncating cell', output['text'][0])
                output['text'] = output['text'][:50]
                output['text'].append('...\n')
                output['text'].append(f'[Long output - {ntrunc} '
                                      'more lines hidden]\n')
                output['text'].append('...\n')

    with open(outfile, 'wt') as f:
        f.write(json.dumps(notebook, indent=2))


def main():

    # get FUNPACK root dir from funpack/scripts
    basedir  = op.abspath(op.join(op.dirname(__file__), '..', '..'))
    demodir  = op.abspath(op.join(op.dirname(demo.__file__), 'demo'))
    testdir  = op.abspath(op.dirname(tests.__file__))
    mdfile   = op.join(demodir, 'funpack_demonstration.md')
    nbfile   = op.join(demodir, 'funpack_demonstration.ipynb')
    exfile   = op.join(testdir, 'funpack_demonstration_with_outputs.ipynb')
    docfile  = op.join(basedir, 'doc', 'demo.ipynb')

    print(f'Converting {mdfile} to {nbfile}')
    markdown_to_notebook(demodir, mdfile, nbfile)

    print(f'Executing {nbfile}, and saving to {exfile}...')
    execute_notebook(nbfile, exfile)

    print(f'Copying executed notebook into doc/ directory...')
    truncate_long_output_cells(exfile, docfile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# main.py - funpack entry point
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the ``funpack`` entry point. """


import multiprocessing     as mp
import os.path             as op
import                        sys
import                        logging
import                        warnings
import                        datetime
import                        calendar

import pandas              as pd
import pandas.api.types    as pdtypes
import                        threadpoolctl

import funpack
import funpack.util        as util
import funpack.icd10       as icd10
import funpack.config      as config
import funpack.custom      as custom
import funpack.dryrun      as dryrun
import funpack.fileinfo    as fileinfo
import funpack.cleaning    as cleaning
import funpack.datatable   as datatable
import funpack.exporting   as exporting
import funpack.hierarchy   as hierarchy
import funpack.importing   as importing
import funpack.loadtables  as loadtables
import funpack.processing  as processing


log = logging.getLogger(__name__)


def main(argv=None):
    """``funpack`` entry point. """

    # Make sure built in plugins are
    # registered, as they are queried
    # in the command-line help. Set
    # logging to critical until we've
    # parsed command-line args.
    logging.getLogger().setLevel(logging.CRITICAL)
    custom.registerBuiltIns()

    args, argv = config.parseArgsWithConfigFile(argv)
    date = datetime.date.today()

    # Set the number of threads
    # that numpy should use
    threadpoolctl.threadpool_limits(args.num_jobs)

    # Now that args are passed,
    # we can set up logging properly.
    configLogging(args)

    log.info('funpack %s', funpack.__version__)
    log.info('Date: %s (%s)', date.today(), calendar.day_name[date.weekday()])
    log.info('Command-line arguments %s', ' '.join(argv))
    log.debug('Running with the following options')
    for name, val in args.__dict__.items():
        if val is not None:
            val = str(val)
            if len(val) <= 30: log.debug('  %s: %s',    name, val)
            else:              log.debug('  %s: %s...', name, val[:30])

    # Load any custom plugins
    # that have been specified.
    if args.plugin_file is not None:
        for p in args.plugin_file:
            custom.loadPluginFile(p)

    # default output format inferred
    # from output filename or, failing
    # that, tsv
    if args.format is None:
        fmt = op.splitext(args.outfile)[1].lower().strip('.')
        if fmt in ('h5', 'hdf'):
            fmt = 'hdf5'
        if not custom.exists('exporter', fmt):
            fmt = 'tsv'
        args.format = fmt

    # error if any loaders/formats are
    # invalid (we can only perform this
    # check after plugins have been
    # loaded)
    if args.loader is not None:
        for f, l in args.loader.items():
            if not custom.exists('loader', l):
                raise ValueError('Unknown loader {} [{}]'.format(l, f))
    if not custom.exists('exporter', args.format):
        raise ValueError('Unknown output format {}'.format(args.format))
    if args.date_format is not None and \
       not custom.exists('formatter', args.date_format):
        raise ValueError('Unknown date format {}'.format(args.date_format))
    if args.time_format is not None and \
       not custom.exists('formatter', args.time_format):
        raise ValueError('Unknown time format {}'.format(args.time_format))
    if args.var_format is not None:
        for v, f in args.var_format.items():
            if not custom.exists('formatter', f):
                raise ValueError('Unknown formatter {} [{}]'.format(f, v))

    if args.num_jobs > 1:
        log.debug('Running up to %i jobs in parallel', args.num_jobs)

        # Funpack relies on fork() to share
        # the dataset to child processes
        # when parallelising tasks. This
        # is potentially dangerous on macOS
        # - see the "macOS issues" note in
        # README, and the python bug report:
        # https://bugs.python.org/issue33725
        if mp.get_start_method(True) is None:
            mp.set_start_method('fork')

        # The icd10 module maintains information
        # which is potentially shared (read from
        # and written to) by multiple processes,
        # so we use a mp.Manager to handle
        # shared state.
        mgr = mp.Manager()
        icd10.initialise(mgr)
    else:
        mgr = None

    with util.timed(None, log, fmt='Total time: %s (%+iMB)'):

        dtable, unknowns, uncategorised, drop = doImport(args, mgr)

        if args.dry_run:
            dryrun.doDryRun(dtable, unknowns, uncategorised, drop, args)
        else:
            doCleanAndProcess(  dtable, args)
            doExport(           dtable, args)
            doUnknownsExport(   dtable, args, unknowns, uncategorised)
            doICD10Export(              args)
            doDescriptionExport(dtable, args)
            doSummaryExport(    dtable, args)

    return 0


def doImport(args, mgr):
    """Data import stage.

    :arg args: :class:`argparse.Namespace` object containing command line
               arguments
    :arg mgr:  :class:`multiprocessing.Manager` object for parallelisation (may
               be ``None``)

    :returns:  A tuple containing:

                - A :class:`.DataTable` containing the data
                - A sequence of :class:`.Column` objects representing the
                  unknown columns.
                - A sequence of :class:`.Column` objects representing columns
                  which are uncategorised, and have no processing or cleaning
                  rules specified on them.
                - A list of :class:`.Column` objects that were not loaded from
                  each input file.
    """

    # if --remove_duplicates, we append
    # an identifying suffix to the names
    # of columns to be removed. This is
    # then passed through as an exclusion
    # pattern to the importData function
    # via its excludeColnames option.
    if args.remove_duplicates:
        suffix           = importing.REMOVE_DUPLICATE_COLUMN_IDENTIFIER
        renameDuplicates = True
    else:
        suffix           = None
        renameDuplicates = args.rename_duplicates

    finfo = fileinfo.FileInfo(args.infile,
                              indexes=args.index,
                              loaders=args.loader,
                              encodings=args.encoding,
                              renameDuplicates=renameDuplicates,
                              renameSuffix=suffix)

    with util.timed('Table import', log):
        vartable, proctable, cattable, unknowns, uncategorised = \
            loadtables.loadTables(
                finfo,
                args.variable_file,
                args.datacoding_file,
                args.type_file,
                args.processing_file,
                args.category_file,
                noBuiltins=args.no_builtins,
                naValues=args.na_values,
                childValues=args.child_values,
                recoding=args.recoding,
                clean=args.clean,
                typeClean=args.type_clean,
                globalClean=args.global_clean,
                skipProcessing=args.skip_processing,
                prependProcess=args.prepend_process,
                appendProcess=args.append_process)


    subjects, exprs = args.subject
    if suffix is None: excludeColnames = []
    else:              excludeColnames = [suffix]

    # Import data
    with util.timed('Data import', log):
        dtable, drop = importing.importData(
            fileinfo=finfo,
            vartable=vartable,
            proctable=proctable,
            cattable=cattable,
            variables=args.variable,
            colnames=args.column,
            excludeColnames=excludeColnames,
            categories=args.category,
            subjects=subjects,
            subjectExprs=exprs,
            exclude=args.exclude,
            excludeVariables=args.exclude_variable,
            excludeCategories=args.exclude_category,
            trustTypes=args.trust_types,
            mergeAxis=args.merge_axis,
            mergeStrategy=args.merge_strategy,
            indexVisits=args.index_visits,
            dropNaRows=args.drop_na_rows,
            njobs=args.num_jobs,
            mgr=mgr,
            dryrun=args.dry_run)

    # Filter unknown/uncategorised
    # column lists to only contain
    # those that were loaded
    allcols       = dtable.dataColumns
    unknowns      = [c for c in unknowns      if c in allcols]
    uncategorised = [c for c in uncategorised if c in allcols]

    # if it appears that we're doing
    # a full run on a large data set,
    # emit warnings about unknown/
    # uncategorised variables.
    bigrun = any((args.variable_file   is not None,
                  args.datacoding_file is not None,
                  args.processing_file is not None,
                  args.category_file   is not None))

    if bigrun:
        for u in unknowns:
            log.warning('Variable %s [file %s, column %s, assigned '
                        'variable ID %s] is unknown.',
                        u.name, u.datafile, u.index, u.vid)
        for u in uncategorised:
            log.warning('Variable %s [file %s, column %s, assigned '
                        'variable ID %s] is uncategorised.',
                        u.name, u.datafile, u.index, u.vid)

    return dtable, unknowns, uncategorised, drop


def doCleanAndProcess(dtable, args):
    """Data cleaning and processing stage.

    :arg dtable: :class:`.DataTable` containing the data
    :arg args:   :class:`argparse.Namespace` object containing command line
                 arguments
    :arg pool:   :class:`multiprocessing.Pool` object for parallelisation (may
                 be ``None``)
    """

    # Clean data (it times each step individually)
    cleaning.cleanData(
        dtable,
        skipNAInsertion=args.skip_insertna,
        skipCleanFuncs=args.skip_clean_funcs,
        skipChildValues=args.skip_childvalues,
        skipRecoding=args.skip_recoding)

    # Process data
    with util.timed('Data processing', log):
        processing.processData(dtable)


def splitDataTable(dtable, args):
    """Splits the .:class:`DataTable` into separate numeric/non-numeric tables.

    Called by :func:`doExport`.  If the ``--suppress_non_numerics`` and/or
    ``--write_non_numerics`` options are active, non-numeric columns need to
    be separated from numeric columns, and possibly saved to a separate output
    file.

    :arg dtable: :class:`.DataTable` containing the data
    :arg args:   :class:`argparse.Namespace` object containing command line
                 arguments
    :returns:    A list of ``(DataTable, filename)`` tuples, containing the
                 :class:`.DataTable` instances and corresponding output
                 file names.
    """

    # We're outputting one main output file
    # with all columns - no need to split
    if not (args.suppress_non_numerics or args.write_non_numerics):
        return [(dtable, args.outfile)]

    # we need to separate out numeric from
    # non-numeric columns, and potentially
    # create two separate datatables
    dtables = []
    ncols   = []
    nncols  = []

    # we run a single value from each
    # column through formatting in
    # order to determine whether each
    # column is numeric or non-numeric
    for col in dtable.dataColumns:

        idx = dtable[:, col.name].first_valid_index()

        # column is all nan
        if idx is None:
            continue

        # format a subtable containing
        # just the column and the first
        # non-na value
        series = exporting.formatColumn(
            col,
            dtable.subtable([col], [idx]),
            dateFormat=args.date_format,
            timeFormat=args.time_format,
            formatters=args.var_format)[:, col.name]

        # separate accordingly
        if pdtypes.is_numeric_dtype(series): ncols .append(col)
        else:                                nncols.append(col)

    # if suppress, only numeric columns
    # are saved to main output file
    if args.suppress_non_numerics:
        log.debug('Separating out %u / %u numeric columns for export',
                  len(ncols), len(dtable.dataColumns))
        dtables.append((dtable.subtable(ncols), args.outfile))
    else:
        dtables.append((dtable, args.outfile))

    # if write, non-numeric columns
    # are saved to an auxillary file
    if args.write_non_numerics:
        log.debug('Separating out %u / %u non-numeric columns for export',
                  len(nncols), len(dtable.dataColumns))
        dtables.append((dtable.subtable(nncols), args.non_numerics_file))

    return dtables


def doExport(dtable, args):
    """Data export stage.

    :arg dtable: :class:`.DataTable` containing the data
    :arg args:   :class:`argparse.Namespace` object containing command line
                 arguments
    """

    # if ids_only is specified, all we
    # need to do is output the index
    if args.ids_only:
        with open(args.outfile, 'wt') as f:
            for i in dtable.index:
                f.write(f'{i}\n')
        return

    # Otherwise we are exporting the full
    # data set, and things become more
    # complicated...

    # List of data tables and file names
    # for export (we may split the dtable
    # up into two - numeric/non-numeric)
    dtables = splitDataTable(dtable, args)

    # If not parallelising, we export the
    # entire file in one go. Because what's
    # the point in chunked export if we're
    # not parallelising across chunks?
    if args.num_jobs <= 1:
        args.num_rows = len(dtable)

    with util.timed('Data export', log):
        for dtable, outfile in dtables:

            # set the DataTable singleton for
            # shared mem to child processes
            # during export.
            datatable.DataTable.setInstance(dtable)

            exporting.exportData(
                dtable,
                outfile,

                # General export options
                fileFormat=args.format,
                numRows=args.num_rows,
                dropNaRows=args.drop_na_rows,
                dateFormat=args.date_format,
                timeFormat=args.time_format,
                formatters=args.var_format,

                # TSV options
                escapeNewlines=args.escape_newlines,
                sep=args.tsv_sep,
                missingValues=args.tsv_missing_values,

                # HDF5 options
                key=args.hdf5_key,
                style=args.hdf5_style)


def doICD10Export(args):
    """If a ``--icd10_map_file`` has been specified, the ICD10 codes present
    in the data (and their converted values) are saved out to the file.
    """
    if args.icd10_map_file is None:
        return

    with util.timed('ICD10 mapping export', log):
        try:
            ihier = hierarchy.getHierarchyFilePath(name='icd10')
            ihier = hierarchy.loadHierarchyFile(ihier)
            icd10.saveCodes(args.icd10_map_file, ihier)

        except Exception as e:
            log.warning('Failed to export ICD10 mappings: {}'.format(e),
                        exc_info=True)


def doDescriptionExport(dtable, args):
    """If a ``--description_file`` has been specified, a description for every
    column is saved out to the file.
    """
    if args.description_file is None:
        return

    with util.timed('Description export', log):
        cols = dtable.dataColumns

        try:
            with open(args.description_file, 'wt') as f:
                for col in cols:
                    desc = generateDescription(dtable, col)
                    f.write('{}\t{}\n'.format(col.name, desc))

        except Exception as e:
            log.warning('Failed to export descriptions: {}'.format(e),
                        exc_info=True)


def generateDescription(dtable, col):
    """Called by :func:`doDescriptionExport`. Generates and returns a
    suitable description for the given column.

    :arg dtable: :class:`.Datatable` instance
    :arg col:    :class:`.Column` instance
    """
    vartable = dtable.vartable
    desc     = vartable.loc[col.vid, 'Description']

    if pd.isna(desc) or (desc == col.name):
        desc = 'n/a'

    # If metadata has been added to the column,
    # we add it to the description. See the
    # binariseCategorical processing function
    # for an example of this.
    if col.metadata is not None:
        suffix = ' ({})'.format(col.metadata)
    else:
        suffix = ' ({}.{})'.format(col.visit, col.instance)

    return '{}{}'.format(desc, suffix)


def doUnknownsExport(dtable, args, unknowns, uncategorised):
    """If the ``--unknown_vars_file`` argument was used, the unknown/
    unprocessed columns are saved out to a file.

    :arg dtable:        :class:`.DataTable` containing the data
    :arg args:          :class:`argparse.Namespace` object containing command
                        line arguments
    :arg unknowns:      List of :class:`.Column` objects representing the
                        unknown columns.
    :arg uncategorised: A sequence of :class:`.Column` objects representing
                        columns which are uncategorised, and have no processing
                        or cleaning rules specified on them.
    """

    if args.unknown_vars_file is None:
        return

    if len(unknowns) + len(uncategorised) == 0:
        return

    # Save unknown/uncategorised
    # vars list to file columns:
    #  - name      - column name
    #  - file      - originating input file
    #  - class     - unknown or uncategorised
    #  - exported  - whether column passed processing and was exported
    allcols     = list(dtable.dataColumns)
    allunknowns = list(unknowns + uncategorised)

    names       = [u.name            for u in allunknowns]
    files       = [u.datafile        for u in allunknowns]
    classes     = ['unknown'         for u in unknowns] + \
                  ['uncategorised'   for u in uncategorised]
    exported    = [int(u in allcols) for u in allunknowns]
    rows        = ['{}\t{}\t{}\t{}'.format(n, f, c, e)
                   for n, f, c, e in zip(names, files, classes, exported)]

    log.debug('Saving unknown/uncategorised variables to %s',
              args.unknown_vars_file)

    try:
        with open(args.unknown_vars_file, 'wt') as f:
            f.write('name\tfile\tclass\texported\n')
            f.write('\n'.join(rows))

    except Exception as e:
        log.warning('Error saving unknown variables to {}: '
                    '{}'.format(args.unknown_vars_file, e),
                    exc_info=True)


def doSummaryExport(dtable, args):
    """If a ``--summary_file`` has been specified, a summary of the cleaning
    steps that have been applied to each variable are saved out to the file.
    """
    if args.summary_file is None:
        return

    vartable = dtable.vartable
    vids     = sorted(dtable.variables)[1:]
    sumdf    = pd.DataFrame(columns=['NAValues', 'RawLevels', 'NewLevels',
                                     'ParentValues', 'ChildValues',
                                     'Clean', 'Flags'], index=vids)
    sumdf.index.name = 'ID'

    with util.timed('Summary export', log):
        for vid in vids:
            sumdf.at[vid, 'NAValues']     = vartable.at[vid, 'NAValues']
            sumdf.at[vid, 'RawLevels']    = vartable.at[vid, 'RawLevels']
            sumdf.at[vid, 'NewLevels']    = vartable.at[vid, 'NewLevels']
            sumdf.at[vid, 'ParentValues'] = vartable.at[vid, 'ParentValues']
            sumdf.at[vid, 'ChildValues']  = vartable.at[vid, 'ChildValues']

            clean = vartable.at[vid, 'Clean']
            if pd.notna(clean):
                sumdf.at[vid, 'Clean'] = list(clean.values())

            flagstr  = []
            cols     = dtable.columns(vid)
            colflags = {c : dtable.getFlags(c) for c in cols}
            flags    = set.union(*colflags.values())

            for flag in flags:
                if all([flag in colflags[c] for c in cols]):
                    flagstr.append(flag)
                else:
                    names = [c.name for c in cols if flag in colflags[c]]
                    flagstr.append('{} [{}]'.format(flag, ', '.join(names)))

            sumdf.at[vid, 'Flags'] = ';'.join(flagstr)

        sumdf.to_csv(args.summary_file, sep='\t')


def configLogging(args):
    """Configures ``funpack`` logging.

    :arg args: ``argparse.Namespace`` object containing parsed command line
               arguments.
    """

    # Custom log handler which
    # colours messages
    class LogHandler(logging.StreamHandler):

        def emit(self, record):

            levelno = record.levelno

            if   levelno >= logging.WARNING:  colour = '\x1b[31;1m'
            elif levelno >= logging.INFO:     colour = '\x1b[39;1m'
            elif levelno >= logging.DEBUG:    colour = '\x1b[90;1m'
            else:                             colour = ''

            # Reset terminal attributes
            # after each message.
            record.msg = '{}{}\x1b[0m'.format(colour, record.msg)

            return super(LogHandler, self).emit(record)

    logger = logging.getLogger('funpack')
    fmt    = logging.Formatter('%(asctime)s '
                               '%(levelname)8.8s '
                               '%(filename)20.20s '
                               '%(lineno)4d: '
                               '%(funcName)-15.15s - '
                               '%(message)s',
                               '%H:%M:%S')

    if args.log_file is None: handler = LogHandler()
    else:                     handler = logging.FileHandler(args.log_file)

    handler.setFormatter(fmt)
    logger.addHandler(handler)

    # configure verbosity
    if   args.quiet:      loglevel = logging.CRITICAL
    elif args.noisy == 0: loglevel = logging.INFO
    else:                 loglevel = logging.DEBUG

    logging.getLogger('funpack').setLevel(loglevel)

    if args.quiet or args.noisy < 3:
        warnings.filterwarnings('ignore',  module='pandas')
        warnings.filterwarnings('ignore',  module='numpy')
        warnings.filterwarnings('ignore',  module='tables')

    if args.noisy == 1:
        makequiet = ['funpack.expression',
                     'funpack.custom',
                     'funpack.cleaning_functions',
                     'funpack.processing_functions']
    elif args.noisy == 2:
        makequiet = ['funpack.expression',
                     'funpack.custom']
    else:
        makequiet = []

    for mod in makequiet:
        logging.getLogger(mod).setLevel(logging.INFO)


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#
# test_main_non_numeric.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import             warnings
import             os
import textwrap as tw
import os.path  as op
import multiprocessing as mp
from unittest import mock
from io import StringIO

import pytest

from collections import OrderedDict

import pandas as pd
import numpy  as np
import           h5py

from pandas.testing import assert_frame_equal

import funpack.importing      as importing
import funpack.exporting      as exporting
import funpack.exporting_hdf5 as exporting_hdf5
import funpack.exporting_tsv  as exporting_tsv
import funpack.datatable      as datatable
import funpack.custom         as custom
import funpack.util           as util
import funpack.main           as main

from . import (gen_DataTable,
               clear_plugins,
               patch_base_tables,
               patch_logging,
               tempdir,
               gen_test_data,
               gen_tables,
               gen_DataTableFromDataFrame)


@patch_logging
def test_main_write_non_numerics():

    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0,4-0.0
    1,11,a,31,ff
    2,12,b,32,gg
    3,13,c,33,hh
    4,14,d,34,ii
    5,15,e,35,jj
    """).strip()
    expn = tw.dedent("""
    eid\t1-0.0\t3-0.0
    1\t11\t31
    2\t12\t32
    3\t13\t33
    4\t14\t34
    5\t15\t35
    """).strip()
    expnn = tw.dedent("""
    eid\t2-0.0\t4-0.0
    1\ta\tff
    2\tb\tgg
    3\tc\thh
    4\td\tii
    5\te\tjj
    """).strip()

    vt = gen_tables((1, 2, 3, 4),
                    vtypes={1 : 'continuous',
                            2 : 'text',
                            3 : 'continuous',
                            4 : 'text'})[0]
    def converttype(t):
        return t.name
    vt['Type'] = vt['Type'].apply(converttype)

    dtypes = {'1-0.0' : np.float64,
              '2-0.0' : str,
              '3-0.0' : np.float64,
              '4-0.0' : str}

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vt.to_csv('variables.tsv', sep='\t')

        data  = pd.read_csv(StringIO(data),  index_col=0, dtype=dtypes)
        expn  = pd.read_csv(StringIO(expn),  index_col=0, dtype=dtypes,
                            sep='\t')
        expnn = pd.read_csv(StringIO(expnn), index_col=0, dtype=dtypes,
                            sep='\t')

        base = ' -nb -vf variables.tsv -ow out.csv data.txt'

        # default: both numeric and non-
        # numeric columns output to main file
        main.main(base.split())
        df = pd.read_csv('out.csv', index_col=0)
        assert_frame_equal(df, data)

        # --suppress_non_numerics - outfile
        # only contains numeric cols
        main.main(('-esn' + base).split())
        df = pd.read_csv('out.csv', index_col=0)
        assert_frame_equal(df, expn)

        # --write_non_numerics - non-numerics
        # saved to both main and auxillary
        # output files
        main.main(('-wnn' + base).split())
        assert_frame_equal(pd.read_csv('out.csv',              index_col=0), data)
        assert_frame_equal(pd.read_csv('out_non_numerics.csv', index_col=0), expnn)

        # --suppress + --write - non-numerics
        # saved to  auxillary output file
        main.main(('-esn -wnn' + base).split())
        assert_frame_equal(pd.read_csv('out.csv',              index_col=0), expn)
        assert_frame_equal(pd.read_csv('out_non_numerics.csv', index_col=0), expnn)



#!/usr/bin/env python
#
# hierarchy.py - Functions and data structures for working with hierarchical
#                variables.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions and data structures for working with
hierarchical variables.


The :func:`loadHierarchyFile` function will read hierarchy information for
one variable from a text file, and return a :class:`Hierarchy` object.


The :class:`Hierarchy` class allows the hierarchy information about a variable
to be queried.


The hierarchy information for all hierarchical variables in the UKBiobank is
stored in ``funpack/data/hierarchy/`` - these files are available from the
UKBiobank showcase at https://biobank.ctsu.ox.ac.uk/crystal/schema.cgi.
"""


import os.path     as op
import functools   as ft
import dataclasses as dc

import pandas      as pd
import numpy       as np

from typing import Any


HIERARCHY_DATA_NAMES = {
    'icd10' : 19,
    'icd9'  : 87,
    'opcs4' : 240,
    'opcs3' : 259,
}
"""This dictionary contains some UK Biobank hierarchical data codings which
can be looked up by name with the :func:`getHierarchyFilePath` function.
"""


class CircularError(Exception):
    """Error raised by the :meth:`Hierarchy.parents` method in the event
    that a circular relationship is detected in a hierarchy.
    """


@ft.lru_cache()
def loadEncodingTable():
    """Loads the encoding.txt file built into funpack, which contains
    information about all UK Biobank encodings.
    """
    datadir   = op.join(op.dirname(__file__), 'data')
    encodings = op.join(datadir, 'encoding.txt')
    encodings = pd.read_csv(encodings, sep='\t', index_col='encoding_id')
    return encodings


def dataCodingType(coding):
    """Returns a data type suitable for representing values in the given
    encoding.
    """
    encodings = loadEncodingTable()
    typecode  = encodings.loc[coding, 'coded_as']

    # the coded_as column conains a numeric ID
    # for each encoding, denoting its type. The
    # codes are the same as those used in
    # field.txt (see loadtables.loadTableBases).
    dtypes = {
        11 : int,
        21 : int,
        31 : float,
        41 : str,
        51 : str
    }
    return dtypes.get(typecode, str)


def getHierarchyFilePath(dtable=None,
                         vid=None,
                         name=None,
                         coding=None):
    """Return a path to the file containing hierarchy information for the
    specified variable/name. Pass the path to the :func:`loadHierarchyFile`
    function to load it.

    Hierarchy files can be looked up with one of the following methods, in
    order of precedence:

     1. By specifying a data coding (``coding``). This takes precedence.
     2. By specifying a ``name`` which is present in the
        :attr:`HIERARCHY_DATA_NAMES`.
     3. By passing a :class:`.DataTable` (``dtable``) and variable ID (``vid``)

    Te recognised data type names for use with the second method are listed in
    the :attr:`HIERARCHY_DATA_NAMES` dictionary.

    A :exc:`ValueError` is raised if the variable is unknown, or does not
    have a listed data coding.

    The returned path is not guaranteed to exist.

    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg name:   Data coding name
    :arg coding: Data coding ID
    :returns:    A path to the relevant hierarchical coding file.
    """

    if coding is None:
        if name is not None:
            coding = HIERARCHY_DATA_NAMES[name.lower()]
        elif (dtable is not None) and (vid is not None):
            try:
                coding = int(dtable.vartable.loc[vid, 'DataCoding'])
            except Exception:
                raise ValueError('Variable {} is unknown or does not '
                                 'have a data coding'.format(vid))
    if coding is None:
        raise ValueError('A coding, name, or dtable+vid must be specified')

    hierdir = op.join(op.dirname(__file__), 'data', 'hierarchy')
    return op.join(hierdir, 'coding{}.tsv'.format(int(coding)))


@ft.lru_cache()
def loadHierarchyFile(fname):
    """Reads the hierarchy information for a variable from the given file.

    ``fname`` is assumed to refer to a tab-separated file containing
    the following columns:

      - ``coding``:    A variable value (not necessarily unique)
      - ``meaning``:   Description
      - ``node_id``:   Unique numeric identifier for each node
      - ``parent_id``: Identifier of each node's parent

    It is assumed that all codings have a unique ``node_id``. Top-level parent
    nodes (nodes with no parent of their own) often have an ID of 0, although
    this is not assumed.

    :arg fname: File containing hierarchy information for one variable
    :returns:   A :class:`Hierarchy` object.
    """

    # get the encoding ID fromn "coding<id>.tsv",
    # and look up a suitable data type
    encoding  = int(op.basename(fname)[6:-4])
    dtype     = {'coding' : dataCodingType(encoding)}

    # load the hierarchy encoding file, ensuring
    # that the coding values are converted
    data      = pd.read_csv(fname, sep='\t', index_col=False, dtype=dtype)
    codings   = data['coding']   .values
    meanings  = data['meaning']  .values
    nodeIds   = data['node_id']  .values
    parentIds = data['parent_id'].values
    return Hierarchy(nodeIds, parentIds, codings, meanings)


def codeToNumeric(code,
                  name=None,
                  dtable=None,
                  vid=None,
                  coding=None,
                  hier=None):
    """Converts a hierarchical code into a numeric version. See the
    :func:`getHierarchyFilePath` for information on the arguments.

    Some hierarchical codings in the UKB do not have unique coding values for
    parent/non-leaf nodes. If this function is passed such a value, ``np.nan``
    is returned.

    :arg code:   Code to convert
    :arg name:   Data coding name
    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg coding: Data coding ID
    :arg hier:   A :class:`Hierarchy` instance which, if provided, will
                 be used instead of loading one from file using the other
                 arguments.
    """
    # We use the node IDs defined in
    # the hierarchy file as the
    # numeric version of each coding.
    if hier is None:
        hier = getHierarchyFilePath(dtable, vid, name, coding)
        hier = loadHierarchyFile(hier)

    try:
        return int(hier.index(code))
    except KeyError:
        return np.nan


def numericToCode(code, name=None, dtable=None, vid=None, coding=None):
    """Converts a numeric hierarchical code into its original version

    :arg code:   Code to convert
    :arg name:   Data coding name
    :arg dtable: The :class:`.DataTable`
    :arg vid:    Variable ID
    :arg coding: Data coding ID
    """
    hier = getHierarchyFilePath(dtable, vid, name, coding)
    hier = loadHierarchyFile(hier)

    try:
        return hier.coding(int(code))
    except KeyError:
        return 'NaN'


@dc.dataclass
class Node:
    """Represnts a node in a hierarchical encoding. Used by the
    :class:`Hierarchy` class.
    """
    node_id:   int
    parent_id: int
    coding:    Any
    attrs:     dict


class Hierarchy:
    """The ``Hierarchy`` class allows information in a hierarchical variable to
    be queried. The :meth:`parents` method will return all parents in the
    hierarchy for a given value (a.k.a. *coding*), and the :meth:`description`
    method will return the description for a value.

    Additional metadata can be added and retrieved for codings via the
    :meth:`set` and :meth:`get` methods.
    """


    def __init__(self, nodes, parents, codings, descs):
        """Create a ``Hierarchy`` object.

        :arg nodes:   Node IDs. Assumed to be unique.
        :arg parents: Parent IDs for each node.
        :arg codings: Value/coding for each node.
        :arg descs:   Description for each node.
        """

        # Mappings from node ID and from
        # coding value to Node objects.
        # Mappings for unique codings
        # only are retained - non-unique
        # mappings are culled afterwards.
        self.__by_id     = {}
        self.__by_coding = {}
        counts           = {}

        for i, node_id in enumerate(nodes):
            coding    = codings[i]
            parent_id = parents[i]
            attrs     = {'description' : descs[i]}
            node      = Node(node_id, parent_id, coding, attrs)

            self.__by_id[    node_id] = node
            self.__by_coding[coding]  = node
            counts[          coding]  = counts.get(coding, 0) + 1

        for coding, count in counts.items():
            if count > 1:
                self.__by_coding.pop(coding)


    @property
    def codings(self):
        """Return a list of all unique codings in the hierarchy. """
        return list(self.__by_coding.keys())


    def index(self, coding):
        """Return the node ID for the given ``coding``. A ``KeyError``
        is raised for non-uniqwue or invalid codings.
        """
        return self.__by_coding[coding].node_id


    def coding(self, nodeID):
        """Return the coding for the given ``nodeID``. """
        return self.__by_id[nodeID].coding


    def parents(self, coding):
        """Return codings for all parents of the given coding.  A ``KeyError``
        is raised for non-uniqwue or invalid codings.
        """
        pids    = self.parentIDs(self.__by_coding[coding].node_id)
        codings = [self.__by_id[p].coding for p in pids]
        return codings


    def parentIDs(self, nodeID):
        """Return IDs of  all parents of the given node. """

        node       = self.__by_id[nodeID]
        parent     = node
        parent_ids = []
        seen       = set()

        while True:

            parent = self.__by_id.get(parent.parent_id, None)

            # we're at the top of the tree
            if parent is None:
                break

            # circular relationship in tree - should never
            # happen, means the coding file is corrupt
            if parent.node_id in seen:
                raise CircularError(f'{parent.node_id} / {parent.coding}')

            seen      .add(   parent.node_id)
            parent_ids.append(parent.node_id)

        return parent_ids


    def description(self, coding):
        """Return the description for the given coding. """
        return self.get(coding, 'description')


    def get(self, coding, attr):
        """Get the given attribute for the given coding. Returns a ``KeyError``
        for non-unique / invalid codings, or non-existent attributes.
        """
        return self.__by_coding[coding].attrs[attr]


    def set(self, coding, attr, value):
        """Set an attribute for the given coding. Returns a ``KeyError``
        for non-unique / invalid codings.
        """
        self.__by_coding[coding].attrs[attr] = value



#!/usr/bin/env python
#
# test_importing_removeSubjects.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap as tw
import pytest

import pandas as pd
import numpy as np

import funpack.importing.filter as filter

from funpack.datatable import Column
import funpack.fileinfo       as fileinfo
from . import (gen_DataTable, gen_DataTableFromDataFrame,
               tempdir, gen_tables, gen_test_data)


def test_columnsToLoad():

    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0
    1,10,20,30
    2,11,21,31
    3,12,22,32
    4,13,23,33
    """).strip()

    vartable = gen_tables(range(1, 10))[0]

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        finfo = fileinfo.FileInfo('data.txt')
        cols  = finfo.columns('data.txt')

        gc, gd = filter.columnsToLoad(finfo, vartable, None)
        assert gc['data.txt'] == cols
        assert gd             == []

        gc, gd = filter.columnsToLoad(finfo, vartable, [1, 2, 3])
        assert gc['data.txt'] == cols[:4]
        assert gd             == []

        gc, gd = filter.columnsToLoad(finfo, vartable, None, exclude=[2, 3])
        assert gc['data.txt'] == cols[:2]
        assert gd             == cols[2:]

        gc, gd = filter.columnsToLoad(finfo, vartable, [1, 2, 3], exclude=[2, 3])
        assert gc['data.txt'] == cols[:2]
        assert gd             == cols[2:]


def test_filterSubjects():

    data       = np.random.randint(1, 10, (500, 11))
    index      = np.arange(1, 501)
    data[:, 0] = index
    colnames   = ['eid'] + ['{}-0.0'.format(i+1) for i in range(10)]
    cols       = [Column(None, n, i, i) for i, n in enumerate(colnames)]
    df         = pd.DataFrame(data, columns=colnames).set_index('eid')
    data       = data[:, 1:]

    #   - exclude:  list of subjects to *exclude*
    #   - include:  list of subjects to *include*
    #   - exprs:    list of expressions specifying *inclusion*

    # pass through
    got = filter.filterSubjects(df, cols)
    assert np.all(got.index == np.arange(1, 501))
    assert np.all(got == df)

    # include
    got = filter.filterSubjects(df, cols, subjects=np.arange(1, 20))
    assert np.all(got.index == np.arange(1, 20))
    assert np.all(got == df.loc[np.arange(1, 20), :])

    # exclude
    got = filter.filterSubjects(df, cols, exclude=[1, 2, 3])
    assert np.all(got.index == np.arange(4, 501))
    assert np.all(got == df.loc[np.arange(4, 501), :])

    # expr
    mask = data[:, 0] > 5
    got = filter.filterSubjects(df, cols, subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # include + exclude
    got = filter.filterSubjects(
        df, cols, subjects=np.arange(1, 20), exclude=[1, 2, 3])
    assert np.all(got.index == np.arange(4, 20))
    assert np.all(got == df.loc[np.arange(4, 20), :])

    # include + expr
    mask       = data[:, 0] > 5
    mask[100:] = 0
    got = filter.filterSubjects(
        df, cols, subjects=np.arange(1, 101), subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # expr + exclude
    mask     = data[:, 0] > 5
    mask[:3] = 0
    got = filter.filterSubjects(
        df, cols, subjectExprs=['v1 > 5'], exclude=[1, 2, 3])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # include + expr + exclude
    mask          = data[:, 0] > 5
    mask[300:]    = 0
    mask[75:200]  = 0
    got = filter.filterSubjects(
        df, cols,
        subjects=np.arange(1, 301),
        exclude=np.arange(76, 201),
        subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # multiple expressions
    got = filter.filterSubjects(
        df, cols, subjectExprs=['v1 > 5', 'v2 == 9'])
    mask = (data[:, 0] > 5) | (data[:, 1] == 9)
    assert np.all(got.index == np.where(mask)[0] + 1)
    assert np.all(got == data[mask, :])


def test_filterSubjects_multiple_columns():

    def gendata():
        colnames   = ['eid', '1-0.0', '1-1.0', '1-2.0', '2-0.0', '2-1.0']
        variables  = [0,      1,       1,       1,       2,       2]
        cols       = [Column(None, n, i, v) for i, (n, v) in
                      enumerate(zip(colnames, variables))]
        data       = np.random.randint(1, 10, (6, 500))
        data[0, :] = np.arange(1, 501)
        df         = pd.DataFrame({c : d for c, d in zip(colnames, data)})
        df         = df.set_index('eid')
        data       = data[1:, :].T
        return df, cols, data

    def all(s): return s.all(axis=1)
    def any(s): return s.any(axis=1)

    # combine vars with ncolumn
    # mismatch - should be ORed
    # within var, then combined
    df, cols, data = gendata()
    exprs = ['v1 > 2 && v2 < 7']
    exp   = any(data[:, :3] > 2) & any(data[:, 3:] < 7)
    got   = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # combine columns within var
    df, cols, data = gendata()
    exprs = ['all(v1 > 2) && any(v2 < 7)']
    exp = all(data[:, :3] > 2) & any(data[:, 3:] < 7)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # no combining columns - should
    # default to any
    df, cols, data = gendata()
    exprs = ['v1 > 6']
    exp = any(data[:, :3] > 6)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # multipler expressions - ORed together
    df, cols, data = gendata()
    exprs = ['v1 > 6', 'v2 < 4']
    exp = any(data[:, :3] > 6) | any(data[:, 3:] < 4)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()


def test_filter_subjects_column_not_present():
    data       = np.random.randint(1, 10, (500, 11))
    index      = np.arange(1, 501)
    data[:, 0] = index
    colnames   = ['eid'] + ['{}-0.0'.format(i+1) for i in range(10)]
    cols       = [Column(None, n, i, i) for i, n in enumerate(colnames)]
    df         = pd.DataFrame(data, columns=colnames).set_index('eid')

    # expression is skipped if any variable is not present
    got1       = filter.filterSubjects(df, cols, subjectExprs=['v11 > 0'])
    got2       = filter.filterSubjects(df, cols, subjectExprs=['v9  < 0 || '
                                                               'v11 > 0'])

    assert (got1 == df).all().all()
    assert (got2 == df).all().all()



#!/usr/bin/env python
#
# custom.py - Custom plugins for funpack.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides infrastructure for registering and accessing
``funpack`` plugins - custom functions for sniffing and loading data files,
and for cleaning and processing.


The following plugin types currently exist:


  +-------------------+-------------------------------------------------------+
  | Plugin type       |                                                       |
  +-------------------+-------------------------------------------------------|
  | ``sniffer``       | Return information about the columns in a file        |
  | ``loader``        | Load data from a file                                 |
  | ``cleaner``       | Run a cleaning function on a single column            |
  | ``processor``     | Run a processing function on one or more data columns |
  | ``metaproc``      | Run a function on a :class:`.Column` ``metadata``     |
  |                   | value                                                 |
  | ``formatter``     | Format a column for output                            |
  | ``exporter``      | Export the processed data set                         |
  +-------------------+-------------------------------------------------------+


To ensure that the ``funpack`` command line help and generated documentation
is nicely formatted, all plugin functions should have a docstring of the
form::

    \"\"\"functionSignature(args)
    Short description of function on a single line.

    Extended description
    \"\"\"


.. autosummary::
   :nosignatures:

   clearRegistry
   loadPluginFile
   registerPlugin
   registerBuiltIns
   listPlugins
   exists
   get
   args
   run
"""


import importlib.util as imputil
import os.path        as op
import functools      as ft
import                   re
import                   sys
import                   logging
import                   importlib
import                   collections


log = logging.getLogger(__name__)


PLUGIN_TYPES = ['loader',
                'sniffer',
                'formatter',
                'cleaner',
                'processor',
                'metaproc',
                'exporter']


class PluginRegistry:
    """The ``PluginRegistry`` keeps track of, and provides access to all
    registered plugins.
    """

    def __init__(self):
        """Create a ``PluginRegistry``. """

        self.clearRegistry()


    def clearRegistry(self):
        """Clears and resets the contents of the ``PluginRegistry``. """
        self.__plugins    = collections.OrderedDict()
        self.__pluginArgs = {}

        for pt in PLUGIN_TYPES:

            self.__plugins[   pt] = collections.OrderedDict()
            self.__pluginArgs[pt] = {}

            decorator = ft.partial(self.registerPlugin, pt)
            runner    = ft.partial(self.run,            pt)
            dname     = pt
            rname     = 'run{}'.format(pt.capitalize())

            setattr(self, dname, decorator)
            setattr(self, rname, runner)


    def registerPlugin(self, pluginType, pluginName=None, **kwargs):
        """Decorator to register a plugin. If name is not provided, the
        name of the decorated function is used.
        """

        if pluginType not in PLUGIN_TYPES:
            raise ValueError('Unsupported plugin type: {}'.format(pluginType))

        if pluginName in self.__plugins[pluginType]:
            log.warning('Overwriting existing {}: {}'.format(
                pluginType, pluginName))

        def wrapper(func):
            name = pluginName
            if name is None:
                name = func.__name__

            log.debug('Registering custom %s function: %s', pluginType, name)

            self.__plugins[   pluginType][name] = func
            self.__pluginArgs[pluginType][name] = kwargs.copy()
            return func

        return wrapper


    def loadPluginFile(self, filename):
        """Loads the given file, assumed to be a Python module containing funpack
        plugin functions.
        """

        # If file does not exist, assume
        # it is a direct reference to a
        # file in the plugins directory.
        if not op.exists(filename):
            if not filename.endswith('.py'):
                filename = filename + '.py'
            moddir   = op.dirname(op.abspath(__file__))
            filename = op.join(moddir, 'plugins', filename)

        filename = op.abspath(filename)
        name     = re.sub('[^a-zA-Z]', '_', filename)

        log.debug('Loading custom plugin: %s [%s]', name, filename)

        spec = imputil.spec_from_file_location(name, filename)
        mod  = imputil.module_from_spec(spec)
        spec.loader.exec_module(mod)

        # Make sure the plugin
        # contents can be pickled
        sys.modules[name] = mod


    def listPlugins(self, pluginType):
        """List plugins of the specified type. """
        return list(self.__plugins.get(pluginType, []))


    def exists(self, pluginType, pluginName):
        """Returns ``True`` if the given plugin exists, ``False`` otherwise.
        """
        return pluginName in self.__plugins.get(pluginType, [])


    def get(self, pluginType, pluginName):
        """Return a reference to the specified plugin. """
        return self.__plugins[pluginType][pluginName]


    def args(self, pluginType, pluginName):
        """Return a dict containing any arguments for the specified plugin.
        """
        return self.__pluginArgs[pluginType][pluginName]


    def run(self, pluginType, pluginName, *args, **kwargs):
        """Run the specified plugin. """
        log.debug('Calling %s plugin function: %s', pluginType, pluginName)
        return self.__plugins[pluginType][pluginName](*args, **kwargs)


def registerBuiltIns():
    """Ensures that all built-in plugins are in the registry. """

    firstTime = len(registry.listPlugins('cleaner')) == 0

    import funpack.exporting            as ue
    import funpack.exporting_hdf5       as ueh5
    import funpack.exporting_tsv        as uet
    import funpack.cleaning_functions   as cf
    import funpack.processing_functions as pf
    import funpack.metaproc_functions   as mf

    if firstTime:
        log.setLevel(logging.CRITICAL)

    importlib.reload(ue)
    importlib.reload(ueh5)
    importlib.reload(uet)
    importlib.reload(cf)
    importlib.reload(pf)
    importlib.reload(mf)

    if firstTime:
        log.setLevel(logging.NOTSET)


registry       = PluginRegistry()
clearRegistry  = registry.clearRegistry
loadPluginFile = registry.loadPluginFile
registerPlugin = registry.registerPlugin
listPlugins    = registry.listPlugins
exists         = registry.exists
get            = registry.get
args           = registry.args
run            = registry.run


for pt in PLUGIN_TYPES:
    decorator = pt
    runner    = 'run{}'.format(pt.capitalize())
    setattr(sys.modules[__name__], decorator, getattr(registry, decorator))
    setattr(sys.modules[__name__], runner,    getattr(registry, runner))



#!/usr/bin/env python
#
# datatable.py - The DataTable class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DataTable` class, a container
class which holds a reference to the loaded data.
"""


import itertools             as it
import multiprocessing       as mp
import multiprocessing.dummy as mpd
import                          random
import                          string
import                          logging
import                          contextlib
import                          collections
import collections.abc       as abc

import pandas                as pd

import funpack.util as util


log = logging.getLogger(__name__)


AUTO_VARIABLE_ID = 5000000
"""Starting variable ID to use for unknown data. Automatically generated
variable IDs really shoulld not conflict with actual UKB variable IDs.
"""


MODIFIED_COLUMN =  ''.join(random.choices(string.ascii_letters, k=20))
"""Flag used internally by the :class:`DataTable` when merging subtables.
"""


class Column:
    """The ``Column`` is a simple container class containing metadata
    about a single column in a data file.

    See the :func:`.parseColumnName` function for important information
    about column naming conventions in the UK BioBank.

    A fundamental assumption made throughout much of the ``funpack`` code-base
    is that columns with a variable ID (``vid``) equal to 0 are index columns
    (e.g. subject ID).

    A ``Column`` object has the following attributes:

     - ``datafile``: Input file that the column originates from (``None`` for
                     generated columns).

     - ``name``:     Column name.

     - ``origname``: Column name as it appears in the input/output file(s)

     - ``index``:    Location of the column in the input/output file(s)

     - ``vid``:      Variable ID

     - ``visit``:    Visit number

     - ``instance``: Instance number/code

     - ``metadata``: Metadata which may be used to generate descriptive
                     information about the column

     - ``basevid``: ID of the variable that the data in this column was
                     derived/generated from (and has the same data type as),
                     if not the original ``vid``.  For example, see the
                     ``take`` argument to
                     :func:`.processing_functions.binariseCategorical`.

     - ``fillval``:  Fill value to use for missing values, instead of the
                     global default (e.g. as specified by
                     ``--tsv_missing_values``).

    """
    def __init__(self,
                 datafile,
                 name,
                 index,
                 vid=None,
                 visit=0,
                 instance=0,
                 metadata=None,
                 basevid=None,
                 fillval=None,
                 origname=None):

        if basevid is None:
            basevid = vid

        if origname is None:
            origname = name

        self.datafile  = datafile
        self.name      = name
        self.origname  = origname
        self.index     = index
        self.vid       = vid
        self.visit     = visit
        self.instance  = instance
        self.metadata  = metadata
        self.basevid   = basevid
        self.fillval   = fillval


    def __str__(self):
        return 'Column({}, {} -> {}, {}, {}, {}, {}, {}, {})'.format(
            self.datafile,
            self.name,
            self.origname,
            self.index,
            self.vid,
            self.visit,
            self.instance,
            self.basevid,
            self.fillval)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return (isinstance(other, Column)       and
                self.datafile == other.datafile and
                self.name     == other.name     and
                self.origname == other.origname and
                self.index    == other.index    and
                self.vid      == other.vid      and
                self.visit    == other.visit    and
                self.instance == other.instance and
                self.basevid  == other.basevid  and
                self.fillval  == other.fillval)



class DataTable(util.Singleton):
    """The ``DataTable`` is a simple container class.

    It keeps references to the variable and processing tables, and the data
    itself. The :func:`importData` function creates and returns a
    ``DataTable``.

    A ``DataTable`` has the following attributes and helper methods:

    .. autosummary::
       :nosignatures:

       pool
       manager
       parallel
       vartable
       proctable
       cattable
       columns
       allColumns
       present
       visits
       instances
       variables


    Data should be accessed/modified via these methods:

    .. autosummary::
       :nosignatures:

       index
       __getitem__
       __setitem__


    Columns can be added/removed, and rows removed, via these methods:

    .. autosummary::
       :nosignatures:

       maskSubjects
       addColumns
       removeColumns

    Columns can be "flagged" with metadata labels via the :meth:`addFlag`
    method. All of the flags on a column can be retrieved via the
    :meth:`getFlags` method.


    The :meth:`subtable` method can be used to generate a replica
    ``DataTable`` with a specific subset of columns. It is intended for
    parallelisation, so that child processes are given a view of only the
    columns that are relevant to them, and as little copying between processes
    as possible takes place. The :meth:`subtable` and :meth:`merge` methods
    are intended to be used like so:

      1. Create subtables which only contain data for specific columns::

             cols      = ['1-0.0', '2-0.0', '3-0.0']
             subtables = [dtable.subtable([c]) for c in cols]

      2. Use multiprocessing to perform parallel processing on each column::

             def mytask(dtable, col):
                 dtable[:, col] += 5
                 return dtable

             with dtable.pool() as pool:
                 subtables = pool.starmap(mytask, zip(subtables, cols))

      3. Merge the results back into the main table::

             dtable.merge(subtables)

    Modifications must occur through the :meth:`DataTable.__setitem__`
    interface, so it can keep track of which columns have been modified.
    Addition or removal of columns or rows on subtables is not supported.
    """


    def __init__(self,
                 data,
                 columns,
                 vartable,
                 proctable,
                 cattable,
                 njobs=1,
                 mgr=None,
                 subtable=False):
        """Create a ``DataTable``.

        :arg data:      ``pandas.DataFrame`` containing the data.
        :arg columns:   List of :class:`.Column` objects, representing the
                        columns that are in the data.
        :arg vartable:  ``pandas.DataFrame`` containing the variable table.
        :arg proctable: ``pandas.DataFrame`` containing the processing table.
        :arg cattable   ``pandas.DataFrame`` containing the category table.
        :arg njobs:     Number of jobs to use for parallelising tasks.
        :arg mgr:       :class:`multiprocessing.Manager` object for
                        parallelisation
        :arg subtable:  For internal use. Used to differentiate between the
                        main ``DataTable``, and child ``DataTable`` objects
                        created via :meth:`subtable`.
        """

        self.__data      = data
        self.__vartable  = vartable
        self.__proctable = proctable
        self.__cattable  = cattable
        self.__njobs     = njobs
        self.__mgr       = mgr
        self.__subtable  = subtable
        self.__flags     = collections.defaultdict(set)

        # The varmap is a dictionary of
        # { vid : [Column] } mappings,
        # and the colmap is a dictionary
        # of { name : Column } mappings
        self.__varmap = collections.OrderedDict()
        self.__colmap = collections.OrderedDict()

        for col in columns:
            self.__colmap[col.name] = col
            if col.vid in self.__varmap: self.__varmap[col.vid].append(col)
            else:                        self.__varmap[col.vid] = [col]


    def __getstate__(self):
        """Returns the state of this :class:`.DataTable` for pickling. """
        return (self.__data,
                self.__vartable,
                self.__proctable,
                self.__cattable,
                self.__varmap,
                self.__colmap,
                self.__subtable,
                self.__flags)


    def __setstate__(self, state):
        """Set the state of this :class:`.DataTable` for unpickling. """

        self.__data      = state[0]
        self.__vartable  = state[1]
        self.__proctable = state[2]
        self.__cattable  = state[3]
        self.__varmap    = state[4]
        self.__colmap    = state[5]
        self.__subtable  = state[6]
        self.__flags     = state[7]
        self.__njobs     = 1
        self.__mgr       = None


    @contextlib.contextmanager
    def pool(self):
        """Return a ``multiprocessing.Pool`` object for performing tasks in
        parallel on the data. If the ``njobs`` argument passed to
        :meth:`__init__` was ``1``, or if this is a ``DataTable`` instance that
        has been unpickled (i.e. instances which are running in a sub-process),
        a ``multiprocessing.dummy.Pool`` instance is created and returned.
        """
        if self.__njobs == 1: Pool = mpd.Pool
        else:                 Pool = mp.Pool

        with Pool(self.__njobs) as pool:
            yield pool
            pool.close()
            pool.join()


    @property
    def isSubtable(self):
        """Returns ``True`` if this ``DataTable`` was created by a
        :meth:`subtable` call.
        """
        return self.__subtable


    @property
    def manager(self):
        """Returns a ``multiprocessing.Manager`` for sharing data between
        processes.
        """
        if self.__mgr is None: return mpd.Manager()
        else:                  return self.__mgr


    @property
    def njobs(self):
        """Returns the number of jobs that will be used by the :attr:`pool`
        for parallelising tasks.
        """
        return self.__njobs


    @property
    def parallel(self):
        """Returns ``True`` if this invocation of ``funpack`` has the ability
        to run tasks in parallel, ``False`` otherwise.
        """
        return self.__njobs > 1


    @property
    def vartable(self):
        """Returns the ``pandas.DataFrame`` containing the variable table. """
        return self.__vartable


    @property
    def proctable(self):
        """Returns the ``pandas.DataFrame`` containing the processing table.
        """
        return self.__proctable


    @property
    def cattable(self):
        """Returns the ``pandas.DataFrame`` containing the category table. """
        return self.__cattable


    @property
    def index(self):
        """Returns the subject indices."""
        return self.__data.index


    @property
    def shape(self):
        """Returns the (nrows, ncolumns) shape of the data."""
        return self.__data.shape


    @property
    def variables(self):
        """Returns a list of all integer variable IDs present in the data.

        The list includes the index variable (which has an id of ``0``).
        """
        return list(self.__varmap.keys())


    @property
    def allColumns(self):
        """Returns a list of all columns present in the data, including
        index columns.
        """
        return list(it.chain(*[self.columns(v) for v in self.variables]))


    @property
    def indexColumns(self):
        """Returns a list of all index columns present in the data. """
        return list(self.columns(0))


    @property
    def dataColumns(self):
        """Returns a list of all non-index columns present in the data. """
        return [c for c in self.allColumns if c.vid != 0]


    def present(self, variable, visit=None, instance=None):
        """Returns ``True`` if the specified variable (and optionally visit/
        instance) is present in the data, ``False`` otherwise.
        """
        try:
            self.columns(variable, visit, instance)
            return True
        except KeyError:
            return False


    def columns(self, variable, visit=None, instance=None):
        """Return the data columns corresponding to the specified ``variable``,
        ``visit`` and ``instance``.

        :arg variable: Integer variable ID
        :arg visit:    Visit number. If ``None``, column names for all visits
                       are returned.
        :arg instance: Instance number. If ``None``, column names for all
                       instances are returned.
        :returns:      A list of :class:`.Column` objects.
        """

        cols = list(self.__varmap[variable])

        if visit is not None:
            cols = [c for c in cols if c.visit == visit]

        if instance is not None:
            cols = [c for c in cols if c.instance == instance]

        return cols


    def visits(self, variable):
        """Returns the visit IDs for the given ``variable``. """
        cols = self.columns(variable)
        return list({c.visit for c in cols})


    def instances(self, variable):
        """Returns the instance IDs for the given ``variable``. """
        cols = self.columns(variable)
        return list({c.instance for c in cols})


    def maskSubjects(self, mask):
        """Remove subjects where ``mask is False``. """
        self.__data = self.__data[mask]


    def removeColumns(self, cols):
        """Remove the columns described by ``cols``.

        :arg cols: Sequence of :class:`Column` objects to remove.
        """

        names = [c.name for c in cols]

        self.__data.drop(names, axis=1, inplace=True)

        for col in cols:

            vcols = self.__varmap[col.vid]
            vcols.remove(col)

            self.__colmap.pop(col.name)

            if len(vcols) > 0: self.__varmap[col.vid] = vcols
            else:              self.__varmap.pop(col.vid)


    def addColumns(self, series, vids=None, kwargs=None):
        """Adds one or more new columns to the data set.

        .. note:: It is assumed that each ``pandas.Series`` object shares
                  the same row indices as this ``DataTable``.

        :arg series: Sequence of ``pandas.Series`` objects containing the
                     new column data to add.

        :arg vids:   Sequence of variables each new column is associated
                     with. If ``None`` (the default), variable IDs are
                     automatically assigned.

        :arg kwargs: Sequence of dicts, one for each series, containing
                     arguments to pass to the :class:`Column` constructor
                     (e.g. ``visit``, ``metadata``, etc).
        """

        import funpack.loadtables as loadtables  # noqa: E501 pylint: disable=import-outside-toplevel

        if vids   is None: vids   = [None] * len(series)
        if kwargs is None: kwargs = [None] * len(series)

        for s in series:
            if s.name in self.__data.columns:
                raise ValueError(
                    'A column with name {} already exists - remove '
                    'it, or assign to it directly'.format(s.name))

        if len(vids) != len(series):
            raise ValueError('length of vids does not match series')
        if len(kwargs) != len(series):
            raise ValueError('length of kwargs does not match series')

        startidx = len(self.__data.columns)
        idxs     = range(startidx, startidx + len(series))

        # if vids are not provided, auto-generate
        # a vid for each column starting from here.
        startvid = max(max(self.variables) + 1, AUTO_VARIABLE_ID)

        # It's much faster to concatenate two
        # DataFrames together than to seuentially
        # concatenate each series to our existing
        # dataframe. We've already checked for
        # column conflicts, and we assume that
        # row indices are aligned.
        log.debug('Merging %u new columns into main data table.', len(series))

        series      = pd.concat(series, axis=1, verify_integrity=False)
        self.__data = pd.concat((self.__data, series),
                                axis=1,
                                copy=False,
                                verify_integrity=False)

        # create a new Column object
        # describing each new column
        for name, idx, vid, kw in zip(series, idxs, vids, kwargs):

            if vid is None:
                vid      = startvid
                startvid = startvid + 1

            if kw is None:
                kw = {}

            col = Column(None, name, idx, vid, **kw)
            self.__colmap[name] = col

            # new column on existing variable.
            # We assume the data type is the
            # same as the existing columns for
            # this variable
            if vid in self.__varmap:
                self.__varmap[vid].append(col)

            # new variable - the addNewVariable
            # function will sort things out
            else:
                dtype = self.__data[name].dtype
                self.__varmap[vid] = [col]
                loadtables.addNewVariable(
                    self.__vartable, vid, name, dtype)


    def getFlags(self, col):
        """Return any flags associated with the specified column.

        :arg col: :class:`Column` object
        :returns: A ``set`` containing any flags associated with ``col``
        """
        return set(self.__flags[col])


    def addFlag(self, col, flag):
        """Adds a flag for the specified column.

        :arg col:  :class:`Column` object
        :arg flag: Flag to add
        """
        self.__flags[col].add(flag)


    def __getitem__(self, slc):
        """Get the specified slice from the data. This method has
        the same interface as the ``pandas.DataFrame.loc`` accessor.
        """
        return self.__data.loc[slc]


    def __setitem__(self, slc, value):
        """Set the specified slice in the data.

        This method suports a limited form of the ``pandas.DataFrame.loc``
        interface. Slices/labels for both rows and columns must be specified,
        and columns may only be specified via a slice, list, tuple, or
        individual label.
        """
        if not isinstance(slc, tuple) or len(slc) != 2:
            raise RuntimeError('DataTable.__setitem__ requires both '
                               'rows and columns to be indexed.')

        self.__data.loc[slc] = value

        # Flag the column as modified. This is
        # detected by the merge method when
        # it merges subtables back in.
        if self.isSubtable:
            cols = slc[1]
            if isinstance(cols, slice):
                cols = self.__data.columns[cols]
            elif not isinstance(cols, (list, tuple)):
                cols = [cols]

            for col in cols:
                self.addFlag(self.__colmap[col], MODIFIED_COLUMN)


    def __len__(self):
        """Returns the number of rows in the data set. """
        return len(self.__data)


    def subtable(self, columns=None, rows=None):
        """Return a new ``DataTable`` which only contains the data for
        the specified ``columns`` and/or ``rows``.

        This method can be used to create a replica ``DataTable`` where the
        underlying ``pandas.DataFrame`` only contains the specified
        columns. It is intended to be used when parallelising tasks, so that
        child processes are given a view of only the relevant columns.

        .. note:: The :meth:`merge` method cannot be used to merge in a
                  subtable that contains a subset of rows.

        :arg columns: List of :class:`Column` objects.
        :arg rows:    Sequence of row indices.
        :returns:     A new :class:`DataTable`.
        """

        if columns is None:
            columns  = self.dataColumns
            colnames = slice(None)
        else:
            colnames = [c.name for c in columns]

        columns = [self.allColumns[0]] + columns

        if rows is None:
            rows = slice(None)

        return DataTable(self.__data.loc[rows, colnames],
                         columns,
                         self.__vartable,
                         self.__proctable,
                         self.__cattable,
                         self.__njobs,
                         subtable=True)


    def merge(self, subtables):
        """Merge the data from the given ``subtables`` into this ``DataTable``.

        It is assumed that the ``subtables`` each contain a sub-set of the
        columns in this ``DataTable``.

        .. note:: The :meth:`merge` method cannot be used to merge in a
                  subtable that contains a subset of rows.

        :arg subtables: A single ``DataTable``, or a sequence of ``DataTable``
                        instances, returned by :meth:`subtable`.
        """

        if not isinstance(subtables, abc.Sequence):
            subtables = [subtables]

        # Gather a list of all subtable dataframes,
        # and a list of all columns to be copied.
        # We only copy modified columns - we assume
        # that all changes to the subtable occurred
        # via DataTable.__setitem__
        subdfs  = []
        subcols = []

        for subtable in subtables:
            subtcols = [c.name for c in subtable.dataColumns
                        if MODIFIED_COLUMN in subtable.getFlags(c)]
            subdfs.append(subtable[:, subtcols])
            subcols.extend(subtcols)

        # if there are columns to merge from
        # any subtable, create a single dataframe
        # containing all of them - quicker than
        # merging them separately
        if len(subcols) > 0:

            log.debug('merging %u subtable dataframes (%u columns)',
                      len(subdfs), len(subcols))

            if len(subdfs) > 1:
                subdf = pd.concat(subdfs,
                                  axis='columns',
                                  verify_integrity=False,
                                  copy=False)
            else:
                subdf = subdfs[0]

            # merge subtable data into the
            # main dataframe, preserving
            # column ordering
            colorder    = self.__data.columns
            self.__data = pd.concat((self.__data.drop(columns=subcols),
                                     subdf.loc[:, subcols]),
                                    axis='columns',
                                    verify_integrity=False,
                                    copy=False)
            self.__data = self.__data[colorder]

        # copy column metadata
        # over from subtables
        for subtable in subtables:
            for subcol in subtable.dataColumns:

                mycol               = self.__colmap[subcol.name]
                myflags             = self.__flags[mycol]
                subflags            = subtable.getFlags(subcol)
                subflags            = subflags.difference((MODIFIED_COLUMN,))
                self.__flags[mycol] = myflags.union(subflags)

                if subcol.metadata is not None:
                    mycol.metadata = subcol.metadata



#!/usr/bin/env python
#
# test_importing_reindex.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import io
import textwrap as tw

import pandas as pd

from funpack.importing import reindex
from funpack.importing import core
from funpack           import fileinfo

from . import gen_tables, gen_columns, tempdir


def test_genReindexedColumns():

    vartable = gen_tables([1, 2, 3])[0]
    cols     = gen_columns(['eid',
                            (1, 0, 0), (1, 1, 0),
                            (2, 0, 0), (2, 1, 0),
                            (2, 0, 1), (2, 1, 1),
                            (3, 0, 0), (3, 1, 0), (3, 2, 0)])

    vartable.loc[:, 'Instancing'] = 1
    got1 = reindex.genReindexedColumns(cols, vartable)
    got2 = reindex.genReindexedColumns(cols, vartable, onlyRealVisits=False)
    vartable.loc[:, 'Instancing'] = 2
    got3 = reindex.genReindexedColumns(cols, vartable)
    vartable.loc[2, 'Instancing'] = 1
    got4 = reindex.genReindexedColumns(cols, vartable)

    assert got1 == (None, None, None)

    exp2cols = gen_columns(['eid', 'visit',
                            (1, None, 0),
                            (2, None, 0),
                            (2, None, 1),
                            (3, None, 0)])
    exp2map = [2, 2, 3, 3, 4, 4, 5, 5, 5]
    exp2map = {cols[i + 1] : exp2cols[ci] for i, ci in enumerate(exp2map)}
    assert got2[0] == exp2cols
    assert got2[1] == exp2map
    assert got2[2] == {0, 1, 2}

    assert got3[0] == exp2cols
    assert got3[1] == exp2map
    assert got3[2] == {0, 1, 2}

    exp4cols = gen_columns(['eid', 'visit',
                            (1, None, 0),
                            (2, 0, 0),
                            (2, 1, 0),
                            (2, 0, 1),
                            (2, 1, 1),
                            (3, None, 0)])
    exp4map = [2, 2, 3, 4, 5, 6, 7, 7, 7]
    exp4map = {cols[i + 1] : exp4cols[ci] for i, ci in enumerate(exp4map)}

    assert got4[0] == exp4cols
    assert got4[1] == exp4map
    assert got4[2] == {0, 1, 2}


def test_reindexByVisit():
    data = tw.dedent("""
    eid,1-0.0,1-1.0,2-0.0,2-1.0,3-0.0,3-1.0,3-0.1
    1,10,20,30,40,50,60,70
    2,11,21,31,41,51,61,71
    3,12,22,32,42,52,62,72
    4,13,23,33,43,53,63,73
    5,14,24,34,44,54,64,74
    6,15,25,35,45,55,65,75
    """).strip()
    vartable = gen_tables([1, 2, 3])[0]
    oldcols  = gen_columns(['eid',
                            (1, 0, 0), (1, 1, 0),
                            (2, 0, 0), (2, 1, 0),
                            (3, 0, 0), (3, 1, 0),
                            (3, 0, 1)])

    olddf   = pd.read_csv(io.StringIO(data), sep=',', index_col=0)
    newcols = reindex.genReindexedColumns(oldcols, vartable)
    newdf   = reindex.reindexByVisit(olddf, oldcols, *newcols)

    for c in oldcols[1:]:
        newname = '{}.{}'.format(c.vid, c.instance)
        slc     = ((slice(None), c.visit), newname)
        oldvals = newdf.loc[slc]
        newvals = olddf.loc[:, c.name]

        assert (oldvals.index.get_level_values(0) == newvals.index).all()

        assert (oldvals.values == newvals.values).all()


def test_importData_indexVisits():
    data = tw.dedent("""
    eid,1-0.0,1-1.0,2-0.0,2-1.0,3-0.0,3-1.0,3-0.1
    1,10,20,30,40,50,60,70
    2,11,21,31,41,51,61,71
    3,12,22,32,42,52,62,72
    4,13,23,33,43,53,63,73
    5,14,24,34,44,54,64,74
    6,15,25,35,45,55,65,75
    """).strip()
    vt, pt, ct = gen_tables([1, 2, 3])[:3]
    oldcols = gen_columns(['eid',
                           (1, 0, 0), (1, 1, 0),
                           (2, 0, 0), (2, 1, 0),
                           (3, 0, 0), (3, 1, 0),
                           (3, 0, 1)])
    newcols, oldnewmap = reindex.genReindexedColumns(oldcols, vt)[:2]
    for c in newcols:
        c.datafile = 'data.csv'

    with tempdir():
        with open('data.csv', 'wt') as f:
            f.write(data)

        finfo = fileinfo.FileInfo('data.csv')
        olddf = pd.read_csv('data.csv', sep=',', index_col=0)
        dt = core.importData(finfo,
                             vt, pt, ct,
                             indexVisits=True)[0]

        assert dt.allColumns == newcols

        for oldcol, newcol in oldnewmap.items():
            old = olddf.loc[:,                           oldcol.name]
            new = dt[       (slice(None), oldcol.visit), newcol.name]
            assert (old.values == new.values).all()



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import itertools     as it
import functools     as ft
import os.path       as op
import unittest.mock as mock
import                  os
import                  shutil
import                  logging
import                  tempfile
import                  contextlib
import                  collections

import datetime

import numpy as np
import pandas as pd

import funpack.util       as util
import funpack.custom     as custom
import funpack.fileinfo   as fileinfo
import funpack.loadtables as loadtables
import funpack.datatable  as datatable


def patch_logging(func):

    log                   = mock.MagicMock()
    log.getEffectiveLevel = lambda : logging.INFO

    def wrapper(*a, **kwa):
        targets = [
            'funpack.cleaning.log',
            'funpack.cleaning_functions.log',
            'funpack.config.log',
            'funpack.custom.log',
            'funpack.datatable.log',
            'funpack.exporting.log',
            'funpack.exporting_hdf5.log',
            'funpack.exporting_tsv.log',
            'funpack.expression.log',
            'funpack.fileinfo.log',
            'funpack.importing.core.log',
            'funpack.importing.filter.log',
            'funpack.importing.reindex.log',
            'funpack.loadtables.log',
            'funpack.main.log',
            'funpack.merging.log',
            'funpack.processing.log',
            'funpack.processing_functions.log',
            'funpack.processing_functions_core.log',
            'funpack.util.log',
            'logging.getLogger',
        ]
        patches = [mock.patch(t, log) for t in targets]
        [p.start() for p in patches]
        try:
            return func(*a, **kwa)
        finally:
            [p.stop() for p in patches]
    return ft.update_wrapper(wrapper, func)


@contextlib.contextmanager
def patch_base_tables():
    with mock.patch('funpack.loadtables.loadTableBases',
                    return_value=(None, None)):
        yield


def clear_plugins(func):
    def wrapper(*a, **kwa):
        result = func(*a, **kwa)
        custom.clearRegistry()
        return result

    return ft.update_wrapper(wrapper, func)


@contextlib.contextmanager
def tempdir(root=None, changeto=True):

    testdir = tempfile.mkdtemp(dir=root)
    prevdir = os.getcwd()
    try:

        if changeto:
            os.chdir(testdir)
        yield testdir

    finally:
        if changeto:
            os.chdir(prevdir)
        shutil.rmtree(testdir)


def gen_test_data(num_vars,
                  num_subjs,
                  out_file,
                  max_visits=1,
                  max_instances=1,
                  start_var=1,
                  start_subj=1,
                  sep='\t',
                  ctypes=None,
                  missprop=0,
                  names=None,
                  min_visits=1):

    if ctypes is None:
        ctypes = {}


    varids = []
    cols   = []

    for varid in range(start_var, num_vars + start_var):

        nvisits    = np.random.randint(min_visits, max_visits    + 1)
        ninstances = np.random.randint(1,          max_instances + 1)

        for visit, instance in it.product(range(nvisits), range(ninstances)):
            cols.append(util.generateColumnName(varid, visit, instance))
            varids.append(varid)

    # subject IDs
    data = pd.DataFrame(index=range(start_subj, num_subjs + start_subj))
    data.index.name = 'eid'

    for varid, col in zip(varids, cols):
        ctype = ctypes.get(varid, 'float')

        if ctype == 'int':
            coldata = np.random.randint(1, 100, num_subjs)
        elif ctype == 'float':
            coldata = np.random.randint(1, 100, num_subjs)
        elif ctype == 'date':
            ys = np.random.randint(2000, 2019, num_subjs)
            ms = np.random.randint(1,    13,   num_subjs)
            ds = np.random.randint(1,    28,   num_subjs)
            coldata = [datetime.date(y, m, d) for y, m, d in zip(ys, ms, ds)]
        elif ctype == 'time':
            hs = np.random.randint(0,    23, num_subjs)
            ms = np.random.randint(0,    59, num_subjs)
            ss = np.random.randint(0,    59, num_subjs)
            coldata = [datetime.time(h, m, s) for h, m, s in zip(hs, ms, ss)]
        elif ctype == 'datetime':
            ys  = np.random.randint(2000, 2019, num_subjs)
            mos = np.random.randint(1,    13,   num_subjs)
            ds  = np.random.randint(1,    28,   num_subjs)
            hs  = np.random.randint(0,    23,   num_subjs)
            mis = np.random.randint(0,    59,   num_subjs)
            ss  = np.random.randint(0,    59,   num_subjs)
            coldata = [datetime.datetime(y, mo, d, h, mi, s)
                       for y, mo, d, h, mi, s in zip(ys, mos, ds, hs, mis, ss)]

        data[col] = coldata

        if missprop > 0:
            missing = np.random.choice(data.index,
                                       int(round(missprop * num_subjs)))
            data.loc[missing, col] = np.nan

    if names is None:
        names = True
    data.to_csv(out_file, sep, header=names)


table_headers  = {

    'variables'   : 'ID\tType\tInternalType\tDescription\tDataCoding\tInstancing\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean',  # noqa
    'datacodings' : 'ID\tNAValues\tRawLevels\tNewLevels',
    'categories'  : 'ID\tCategory\tVariables',
    'types'       : 'Type\tClean',
    'processing'  : 'Variable\tProcess'
}

table_templates = {
    'variables'   : '{variable}\t{type}\t\t\t\t2\t\t\t\t\t\t',
    'datacodings' : '',
    'categories'  : '',
    'types'       : '',
    'processing'  : '',
}

def gen_tables(variables, vtypes=None, datafiles=None):

    if vtypes is None:
        vtypes = {}

    if datafiles is not None:
        datafiles = [op.abspath(f) for f in datafiles]

    with tempdir():

        if datafiles is None:
            with open('datafile.txt', 'wt') as f:
                colnames = ['eid'] + ['{}-0.0'.format(v) for v in variables]
                f.write('\t'.join(colnames))
            datafiles = ['datafile.txt']

        for table in ['variables',
                      'datacodings',
                      'categories',
                      'types',
                      'processing']:

            fname = '{}.tsv'.format(table)
            hdr   = table_headers[  table]
            tmpl  = table_templates[table]

            with open(fname, 'wt') as f:
                f.write(hdr + '\n')

                if table == 'variables':
                    for v in variables:

                        vtype = str(vtypes.get(v, 'continuous'))

                        f.write(tmpl.format(variable=v, type=vtype) + '\n')

        vt, pt, ct, uk, up = loadtables.loadTables(
            fileinfo.FileInfo(datafiles),
            ['variables.tsv'],
            ['datacodings.tsv'],
            'types.tsv',
            'processing.tsv',
            'categories.tsv',
            noBuiltins=True)
        return vt, pt, ct, uk


# vvis == [(vid, visit, instance), ...]
def gen_columns(vvis):
    cols = []
    for i, vvi in enumerate(vvis):
        if isinstance(vvi, str):
            cols.append(datatable.Column(None, vvi, i, 0))
            continue
        vid, visit, instance = vvi
        if visit is None:
            name = '{}.{}'.format(vid, instance)
            visit = 0
        else:
            name = '{}-{}.{}'.format(vid, visit, instance)
        cols.append(datatable.Column(None, name, i, vid, visit, instance))
    return cols


def gen_DataTable(cols, *a, **kwa):

    nsubjs    = len(cols[0])
    variables = range(0, len(cols) + 1)
    colnames  = ['eid'] + ['{}-0.0'.format(v) for v in variables[1:]]

    columns = collections.OrderedDict(zip(colnames[1:], cols))

    data = pd.DataFrame(columns)
    data['eid'] = np.arange(1, nsubjs + 1)
    data.set_index('eid', inplace=True)

    return gen_DataTableFromDataFrame(data, *a, **kwa)


def gen_DataTableFromDataFrame(df, tables=None, njobs=1, variables=None):

    if variables is None:
        variables = list(range(1, len(df.columns) + 1))

    colobjs   = [datatable.Column(None, df.index.name, 0, 0, 0, 0)] + \
                [datatable.Column(None, n, v, v, 0, 0)
                 for v, n in zip(variables, df.columns)]

    if tables is None:
        variables = list(set(variables))
        vartable, proctable, cattable, uvs = gen_tables(variables)
    else:
        vartable, proctable, cattable = tables

    return datatable.DataTable(df, colobjs, vartable, proctable, cattable,
                               njobs=njobs)



#!/usr/bin/env python
#
# test_demo.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import time
import json
import glob
import shutil
import sys
import os
import signal
import os.path as op
import itertools as it
import subprocess as sp
import multiprocessing as mp

from unittest import mock

import jinja2 as j2
import numpy as np

import notebook.notebookapp as notebookapp
import funpack.scripts.demo as ukbdemo


from . import tempdir


def test_demo():

    def nbfunc():
        # add some extra args for running within docker
        if op.exists('/.dockerenv'): args = ['--allow-root', '--ip=0.0.0.0']
        else:                        args = None

        with mock.patch('sys.argv', ['fmrib_unpack_demo']):
            ukbdemo.main(args)

    nbproc = mp.Process(target=nbfunc)
    nbproc.start()
    time.sleep(3)

    servers = notebookapp.list_running_servers()

    assert sum([s['pid'] == nbproc.pid for s in servers]) == 1
    for s in servers:
        if s['pid'] == nbproc.pid:
            notebookapp.shutdown_server(s)
            break

    os.kill(nbproc.pid, signal.SIGKILL)
    nbproc.join(0.5)
    assert not nbproc.is_alive()


def create_funpack_entrypoint(path):
    with open(path, 'wt') as f:
        f.write('#!/usr/bin/env bash\n')
        f.write('{} -m funpack "$@"\n'.format(sys.executable))
    os.chmod(path, 0o755)


def test_demo_commands():

    def eval_cmd(cmd, out, env):

        # TODO extract all funpack calls, and turn
        #      them into funpack.main function calls.
        result = sp.run(['bash', cmd], stdout=sp.PIPE, env=env)

        with open(out, 'rt') as f:
            out = f.read()
        assert result.returncode              == 0
        assert result.stdout.decode().strip() == out.strip()

    # Issues with pytest-cov 3.0.0 / coveragepy 6.2
    # cause coverage to spit out error information
    # in subprocess-called invocations. Disabling for
    # the time being.
    env = os.environ.copy()
    for k in list(env.keys()):
        if 'COVERAGE' in k or k.startswith('COV_'):
            env.pop(k)

    with tempdir() as td:
        td = op.realpath(td)
        gen_demo_tests(rundir=td)
        create_funpack_entrypoint(op.join(td, 'funpack'))
        create_funpack_entrypoint(op.join(td, 'fmrib_unpack'))

        env['PATH'] = op.pathsep.join((td, env['PATH']))

        demodir = op.join(op.dirname(__file__), '..', 'scripts', 'demo')

        for fname in glob.glob(op.join(demodir, '*.*')):
            shutil.copy(fname, td)

        commands = list(sorted(glob.glob('*_command.txt')))
        outputs  = list(sorted(glob.glob('*_output.txt')))

        for i, (cmd, out) in enumerate(zip(commands, outputs)):

            print('Command', i)
            print(open(cmd).read())
            print()

            eval_cmd(cmd, out, env)


def gen_demo_tests(outdir=None, rundir=None):

    if outdir is None: outdir = os.getcwd()
    if rundir is None: rundir = os.getcwd()

    demofile = op.join(op.dirname(__file__),
                       'funpack_demonstration_with_outputs.ipynb')
    with open(demofile, 'rt') as f:
        content = f.read()

    content = j2.Template(content)
    content = content.render(dir_prefix=rundir, op=op)
    nb      = json.loads(content)
    cells   = nb['cells']

    fname_prefix_format = '{{:0{}d}}'.format(
        int(np.ceil(np.log10(len(cells)))))

    i = 1
    for cell in cells:

        if cell['cell_type'] != 'code':
            continue

        source = cell['source']
        output = cell['outputs']

        # exclude some specific commands
        excludes = ['pygmentize',
                    'ls -l $ukbdir/configs/fmrib']

        if any([exc in line
                for exc, line
                in it.product(excludes, source)]):
            continue

        alloutput = []
        for o in output:
            if o['name'] == 'stdout':
                alloutput += o['text']
        output = alloutput

        cmd_file = '{}_command.txt'.format(
            fname_prefix_format.format(i))
        out_file = '{}_output.txt'.format(
            fname_prefix_format.format(i))

        # Todo abs paths in unknown var file output

        with open(op.join(outdir, cmd_file), 'wt') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write('export PYTHONPATH="{}"\n'.format(
                op.join(op.dirname(__file__), '..', '..')))
            for line in source:
                # The demo uses an alias funpack='fmrib_unpack -q -ow'
                if line.startswith('funpack '):
                    line = line.replace('funpack ', 'funpack -q -ow ')
                    line = '{} -m '.format(sys.executable) + line
                f.write(line)
        with open(op.join(outdir, out_file), 'wt') as f:
            for line in output:
                f.write(line)

        i += 1



#!/usr/bin/env python
#
# reindex.py - Re-indexing a data frame by visit.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions used by the :func:`.core.importData`
function, for re-arranging a data frame that is indexed by subject into
a data frame that is indexed by subject and visit.
"""


import copy
import logging
import collections

import pandas as pd

from .. import datatable


log = logging.getLogger(__name__)


def reindexByVisit(df, oldcols, newcols, oldnewmap, visits):
    """Re-arranges the data so that visits form part of the row
    indices, rather than being stored in separate columns for each variable.

    The :func:`generateReindexedColumns` function is used to create the
    ``newcols``, ``oldnewmap``, and ``visits`` arguments.

    The given dataframe is assumed to be indexed by a single-level
    ``pandas.Index`` This is replaced with a ``pandas.MultiIndex``, where
    the first level corresponds to the original index, and the second level
    to the visit number.

    :arg df:       ``pandas.DataFrame`` containing the data.

    :arg oldcols:  list of :class:`.Column` objects describing the existing
                   data.

    :arg newcols:  List of :class:`.Column` objects decsribing each column in
                   the new adjusted data.

    :arg oldnewmap Dict of ``{old Column : new Column}`` mappings, describing
                   how the old data maps to the new data.

    :returns:      adjusted ``pandas.DataFrame``
    """

    # Create the new two-tierd index
    # comprising existing row IDs
    # and visit numbers
    oldindex    = df.index
    newindex    = pd.MultiIndex.from_product(
        (oldindex, visits), names=(oldindex.name, 'visit'))

    # Create a new empty, but
    # fully labelled, dataframe
    olddf       = df
    newcolnames = [c.name for c in newcols if c.vid != 0]
    newdf       = pd.DataFrame(index=newindex, columns=newcolnames)

    # Loop through column of the
    # old dataframe, copying its
    # data into the corresponding
    # location in the new dataframe.
    for oldcol in oldcols:

        if oldcol.vid == 0:
            continue

        newcol    = oldnewmap[oldcol]
        reindexed = (oldcol != newcol)

        if reindexed: idx = ((slice(None), oldcol.visit), newcol.name)
        else:         idx = ((slice(None), 0),            newcol.name)

        newdf.loc[idx] = olddf.loc[:, oldcol.name].values

    log.debug('Re-indexed for %u visits - %u rows and %u columns '
              're-arranged to %u rows and %u columns', len(visits),
              len(olddf), len(oldcols), len(newdf), len(newcols))

    return newdf


def genReindexedColumns(cols, vartable, onlyRealVisits=True):
    """Figures out how to re-arrange the columns of a data set so that it is
    indexed by row ID and visit.  This function is called by :func:`loadFile`
    when the ``indexVisits`` option is used.

    The ``onlyRealVisits`` argument controls whether the re-indexing is only
    applied to variables which are labelled with `instancing 2
    <https://biobank.ctsu.ox.ac.uk/crystal/instance.cgi?id=2>`_ (Biobank
    assessment centre visit), or is solely based on the ``instance`` component
    of the column name (see the :func:`.parseColumnName` function for details).

    :arg cols:           list of :class:`.Column` objects dewscribing the
                         existing data.

    :arg vartable:       ``pandas.DataFrame`` containing the variable table

    :arg onlyRealVisits: Re-index all columns according to the ``instance``
                         component of their column names, not just columns
                         associated with variables that follow instancing
                         2.

    :returns:            ``None`` if the file does not contain any data with
                         multiple visits, or a ``tuple`` containing:

                           - list of :class:`.Column` objects describing
                             the re-indexed data set.
                           - Dict of ``{old Column : new Column}`` mappings,
                             describing how the old data maps to the new data.
                           - A ``set`` containing all visit codes in the data.
    """

    oldcols = cols
    vt      = vartable

    # List of Column objects for the
    # new adjusted dataframe. During
    # creation, we index them by
    # (vid, visit, instance), to make
    # the mapping of old columns onto
    # new ones easier
    newcols = collections.OrderedDict()

    # Dict of {old Column : new Column}
    # mappings, so we know how to copy
    # data from the old df into the new
    # df
    oldnewmap = {}

    # All visit codes in the data
    visits = set()

    # figure out which variables should
    # be deemed as having visits, and
    # thus subjected to the re-indexing
    # process.
    def hasVisits(vid):
        return (not onlyRealVisits) or (vt.loc[vid, 'Instancing'] == 2)

    # Generate a list of Column objects
    # describing the new data frame. Start
    # with the index column. Currently
    # only supporting single-column index
    # in the input.
    idxcol            = copy.copy([c for c in oldcols if c.vid == 0][0])
    visitcol          = copy.copy(idxcol)
    visitcol.name     = 'visit'
    visitcol.origname = 'visit'
    newcols[0, 0, 0]  = idxcol
    newcols[0, 1, 0]  = visitcol

    # Now regular (non-index) columns
    for oldcol in oldcols:

        if oldcol.vid == 0:
            continue

        vid      = oldcol.vid
        visit    = oldcol.visit
        instance = oldcol.instance
        hasv     = hasVisits(vid)

        # This variable has visits - all
        # old columns with the same vid
        # and instance will map on to
        # the same new column.
        if hasv:
            visits.add(visit)
            visit = 0

        newcol = newcols.get((vid, visit, instance), None)

        # We've already created a new
        # column which maps to this old
        # column (i.e. from a different
        # visit number).
        if newcol is not None:
            oldnewmap[oldcol] = newcol
            continue

        # if this VID has visits, the
        # new dataframe will contain
        # one column for each instance.
        if hasv:
            newcol = datatable.Column(
                oldcol.datafile,
                name='{}.{}'.format(vid, instance),
                index=None,
                vid=vid,
                visit=0,
                instance=instance)

        # if this VID doesn't have visits,
        # the new dataframe will contain
        # identical columns to the old one.
        else:
            newcol = copy.copy(oldcol)

        newcols[vid, visit, instance] = newcol
        oldnewmap[oldcol]             = newcol

    newcols = list(newcols.values())
    for i, col in enumerate(newcols):
        col.index = i

    if len(visits) == 0:
        return (None, None, None)

    return newcols, oldnewmap, visits



#!/usr/bin/env python
#
# dryrun.py - Dry run
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :func:`doDryRun` function, which prints
information about what ``funpack`` would do, but doesn't actually do
anything.
"""


import funpack


def doDryRun(dtable, unknowns, uncategorised, dropped, args):
    """Dry run.

    Prints out information about the cleaning/processing that would be applied.

    :arg dtable:        :class:`.DataTable` containing the data
    :arg unknowns:      List of :class:`.Column` objects representing the
                        unknown columns.
    :arg uncategorised: A sequence of :class:`.Column` objects representing
                        columns which are uncategorised, and have no processing
                        or cleaning rules specified on them.
    :arg dropped:       List of :class:`.Column` objects representing the
                        unknown columns.
    :arg args:          :class:`argparse.Namespace` object containing command
                        line arguments
    """

    variables = dtable.variables

    print('funpack {} dry run'.format(funpack.__version__))
    print()
    print('Input data')
    if len(args.infile) > 1 and  args.merge_strategy == 'naive':
        print('  [Using naive merge strategy - column/'
              'variable count may not be accurate!]')

    print('  Loaded columns:        {}'.format(len(dtable.allColumns)))
    if args.noisy > 0:
        print('\n'.join(['    {}'.format(c.name) for c in dtable.allColumns]))

    print('  Ignored columns:       {}'.format(len(dropped)))
    if args.noisy > 0:
        print('\n'.join(['    {}'.format(c.name) for c in dropped]))

    print('  Unknown columns:       {}'.format(len(unknowns)))
    if args.noisy > 0:
        print('\n'.join(['    {}'.format(c.name) for c in unknowns]))

    print('  Uncategorised columns: {}'.format(len(uncategorised)))
    if args.noisy > 0:
        print('\n'.join(['    {}'.format(c.name) for c in uncategorised]))

    print('  Loaded variables:      {}'.format(len(variables)))
    if args.noisy > 0:
        print('\n'.join(['    {}'.format(v) for v in variables]))
    print()

    print('Cleaning')

    print()
    print('  NA Insertion: {}'.format(not args.skip_insertna))
    if not args.skip_insertna:
        navals = dtable.vartable['NAValues'].dropna().filter(variables)
        if len(navals) == 0:
            print('    No NA value replacements')
        for vid in sorted(navals.index):
            print('    {}: {}'.format(vid, navals[vid]))

    print()
    print('  Cleaning functions: {}'.format(not args.skip_clean_funcs))
    if not args.skip_clean_funcs:
        cleanfuncs = dtable.vartable['Clean'].dropna().filter(variables)
        if len(cleanfuncs) == 0:
            print('    No cleaning functions')
        for vid in sorted(cleanfuncs.index):
            vcfs = ', '.join([str(f) for f in cleanfuncs[vid].values()])
            print('    {}: [{}]'.format(vid, vcfs))

    print()
    print('  Child value replacement: {}'.format(not args.skip_childvalues))
    if not args.skip_childvalues:
        parentvals = dtable.vartable['ParentValues'].dropna().filter(variables)
        childvals  = dtable.vartable['ChildValues'] .dropna().filter(variables)
        if len(parentvals) == 0:
            print('    No replacements')
        for vid in sorted(parentvals.index):
            print('    {}: [{}] -> {}'.format(
                vid,
                ', '.join(map(str, parentvals[vid])),
                childvals[ vid]))

    print('  Categorical recoding: {}'.format(not args.skip_recoding))
    if not args.skip_recoding:
        rawlevels = dtable.vartable['RawLevels'].dropna().filter(variables)
        newlevels = dtable.vartable['NewLevels'].dropna().filter(variables)
        if len(rawlevels) == 0:
            print('    No recodings')
        for vid in sorted(rawlevels.index):
            print('    {}: {} -> {}'.format(vid,
                                            rawlevels[vid],
                                            newlevels[vid]))

    print()
    print('Processing: {}'.format(not args.skip_processing))
    if not args.skip_processing:
        if len(dtable.proctable) == 0:
            print('No processes')
        for idx in sorted(dtable.proctable.index):
            vids = dtable.proctable.loc[idx, 'Variable']
            vps  = dtable.proctable.loc[idx, 'Process']
            vps  = ', '.join([str(p) for p in vps.values()])

            print('  {}: {} -> [{}]'.format(idx + 1, vids, vps))



#!/usr/bin/env python
#
# loadtables.py - Functions which load the variable, data coding, processing,
#                 and category tables used by funpack.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions and logic to load the variable, data coding,
type, processing, and category tables used by funpack.

The variable table is a ``pandas.DataFrame`` which contains metadata about all
UK Biobank variables in the input data files, along with cleaning rules. The
data coding and type tables contain the same information about all UK Biobank
data codings and types - these are merged into the variable table after
being loaded. All of these tables are loaded by the :func:`loadVariableTable`
function.

The processing table contains an ordered list of processing functions to be
applied to the input data.

The category table contains collections of variable groupings; it is used to
allow the user to select these groups by name, rather than having to
use variable IDs.

.. autosummary::
   :nosignatures:

   loadTables
   loadVariableTable
   addNewVariable
   loadProcessingTable
   loadCategoryTable
   categoryVariables
   columnTypes
"""


import itertools as it
import functools as ft
import os.path   as op
import              re
import              logging
import              warnings
import              collections

from typing import Tuple, Sequence, Union, Dict, List, Type, Any, Callable
from typing_extensions import Literal

import numpy  as np
import pandas as pd

import funpack.util       as util
import funpack.fileinfo   as finfo
import funpack.datatable  as datatable
import funpack.processing as processing
import funpack.expression as expression


log = logging.getLogger(__name__)


def convert_type(val : str) -> util.CTYPES:
    """Convert a string containing a UK BioBank type into a numerical
    identifier for that type - see :attr:`funpack.util.CTYPES`.
    """
    valmap = {
        'sequence' :
        util.CTYPES.sequence,
        'integer' :
        util.CTYPES.integer,
        'continuous' :
        util.CTYPES.continuous,
        'categorical (single)' :
        util.CTYPES.categorical_single,
        'categorical (single non-numeric)' :
        util.CTYPES.categorical_single_non_numeric,
        'categorical (multiple)' :
        util.CTYPES.categorical_multiple,
        'categorical (multiple non-numeric)' :
        util.CTYPES.categorical_multiple_non_numeric,
        'time' : util.CTYPES.time,
        'date' :
        util.CTYPES.date,
        'text' :
        util.CTYPES.text,
        'compound' :
        util.CTYPES.compound,
        'unknown' :
        util.CTYPES.unknown,
    }
    return valmap.get(val.lower(), util.CTYPES.unknown)


def convert_dtype(val : str) -> Union[np.dtype, Literal[np.nan]]:
    """Convert a string containing a ``numpy.dtype`` (e.g. ``'float32'``)
    into a ``dtype`` object.
    """

    if val == '':
        return np.nan

    dtype = getattr(np, val, None)

    if dtype not in np.ScalarType:
        raise ValueError('Invalid numpy dtype: {}'.format(dtype))

    return dtype


def convert_comma_sep_text(val : str) -> Union[List[str], Literal[np.nan]]:
    """Convert a string containing comma-separated text into a list. """
    if val.strip() == '':
        return np.nan
    words = val.split(',')
    return [w.strip() for w in words]


def convert_comma_sep_numbers(val : str) -> Union[np.ndarray, Literal[np.nan]]:
    """Convert a string containing comma-separated numbers into a ``numpy``
    array.
    """
    if val.strip() == '':
        return np.nan
    return np.fromstring(val, sep=',', dtype=np.float)


def convert_ParentValues(
        val : str) -> Union[List[expression.Expression], Literal[np.nan]]:
    """Convert a string containing a sequence of comma-separated
    ``ParentValue`` expressions into a sequence of :class:`.Expression`
    objects.
    """
    if val.strip() == '':
        return np.nan
    return [expression.Expression(e) for e in val.split(',')]


def convert_Process_Variable(val : str) -> Tuple[str, List[int]]:
    """Convert a string containing a process variable specification - one of:

      - One or more comma-separated MATLAB-style ``start:stop:step`` ranges,
        indicating that the process is to be applied to the specified variables
        simultaneously.

      - ``'independent,'``, followed by one or more comma-separated
        MATLAB-style ``start:stop:step`` ranges, indicating that the process
        is to be applied to the specified variables independently.

      - ``'all'``, indicating that the process is to be applied to all
        variables simultaneously

      - ``'all_independent'``, indicating that the process is to be applied
        to all variables independently

      - ``'all_except,'``, followed by one or more comma-separated MATLAB-style
        ranges, indicating that the process is to be applied to all variables
        simultaneously, except for the specified variables.

      - ``'all_independent_except,'``, followed by one or more comma-separated
        MATLAB-style ranges, indicating that the process is to be applied to
        all variables independently, except for the specified variables.

    :returns: A tuple containing:

               - The process variable type - one of ``'all'``,
                 ``'all_independent'``, ``'all_except'``,
                 ``'all_independent_except'``, ``'independent'``, or
                 ``'vids'``

               - A list of variable IDs (empty if the process variable type
                 is ``'all'`` or ``'all_independent'``).
    """
    tokens = convert_comma_sep_text(val)
    if len(tokens) == 1 and \
       tokens[0] in ('all', 'all_independent',
                     'all_except', 'all_independent_except'):
        return np.array((tokens[0], []), dtype=object)

    if tokens[0] in ('independent', 'all_except', 'all_independent_except'):
        ptype  = tokens[0]
        tokens = tokens[1:]
    else:
        ptype = 'vids'

    vids = list(it.chain(*[util.parseMatlabRange(t) for t in tokens]))

    return np.array((ptype, vids), dtype=object)


def convert_Process(
        ptype : str,
        val   : str
) -> Dict[str, processing.Process]:
    """Convert a string containing a sequence of comma-separated ``Process`` or
    ``Clean`` expressions into an ``OrderedDict`` of :class:`.Process`
    objects (with the process names used as dictionary keys).
    """
    if val.strip() == '':
        return np.nan

    procs = processing.parseProcesses(val, ptype)

    return collections.OrderedDict([(p.name, p)  for p in procs])


def convert_category_variables(val : str) -> List[int]:
    """Convert a string containing a sequence of comma-separated variable IDs
    or ranges into a list of variable IDs. Variables may be specified as
    integer IDs, or via a MATLAB-style ``start:step:stop`` range. See
    :func:`.util.parseMatlabRange`.
    """

    ranges    = convert_comma_sep_text(val)
    variables = list(it.chain(*[util.parseMatlabRange(r) for r in ranges]))

    return variables


VARTABLE_COLUMNS = [
    'ID',
    'Type',
    'InternalType',
    'Description',
    'DataCoding',
    'Instancing',
    'NAValues',
    'RawLevels',
    'NewLevels',
    'ParentValues',
    'ChildValues',
    'Clean']
"""The columns that must be in a variable table file. """


DCTABLE_COLUMNS = [
    'ID',
    'NAValues',
    'RawLevels',
    'NewLevels']
"""The columns that must be in a datacoding table file. """


TYPETABLE_COLUMNS = [
    'Type',
    'Clean']
"""The columns that must be in a type table file. """


PROCTABLE_COLUMNS = [
    'Variable',
    'Process']
"""The columns that must be in a processing table file. """


CATTABLE_COLUMNS = [
    'ID',
    'Category',
    'Variables']
"""The columns that must be in a category table file. """


VARTABLE_DTYPES = {
    'ID'           : np.uint32,
    'Description'  : object,

    # We can't use an integer for the data
    # coding, because not all variables
    # have a data coding, and pandas uses
    # np.nan to represent missing data.
    'DataCoding'   : np.float32,
    'Instancing'   : np.float32,
    'NAValues'     : object,
    'RawLevels'    : object,
    'NewLevels'    : object,
    'ParentValues' : object,
    'ChildValues'  : object,
    'Clean'        : object,

}
"""Types to use for some columns in the variable table. """


VARTABLE_CONVERTERS = {
    'Type'         : convert_type,
    'InternalType' : convert_dtype,
    'ParentValues' : convert_ParentValues,
    'Clean'        : ft.partial(convert_Process, 'cleaner'),
}
"""Custom converter functinos to use for some columns in the variable
table.
"""


DCTABLE_DTYPES = {
    'ID'         : np.uint32,
    'NAValues'   : object,
    'RawLevels'  : object,
    'NewLevels'  : object,
}
"""Types to use for some columns in the data coding table. """



TYPETABLE_DTYPES = {
    'Type'  : object,
    'Clean' : object,
}
"""Types to use for some columns in the types table. """


TYPETABLE_CONVERTERS = {
    'Type'  : convert_type,
    'Clean' : ft.partial(convert_Process, 'cleaner'),
}
"""Custom converter functinos to use for some columns in the type trable. """



PROCTABLE_CONVERTERS = {
    'Variable' : convert_Process_Variable,
    'Process'  : ft.partial(convert_Process, 'processor'),
}
"""Custom converter functinos to use for some columns in the processing
table.
"""


CATTABLE_DTYPES = {
    'ID' : np.int32,
}
"""Types to use for some columns in the category table. """


CATTABLE_CONVERTERS = {
    'Variables' : convert_category_variables
}
"""Custom converter functinos to use for some columns in the category
table.
"""


IMPLICIT_CATEGORIES = {
    'unknown'       : -1,
    'uncategorised' : -2,
}
"""This dict contains some categories which are automatically/implicitly
added to the category table by the :func:`loadTables` function (via a
call to :func:`addImplicitCategories`).
"""


def loadTables(
        fileinfo : finfo.FileInfo,
        varfiles : Sequence[str] = None,
        dcfiles  : Sequence[str] = None,
        typefile : str           = None,
        procfile : str           = None,
        catfile  : str           = None,
        **kw
) -> Tuple[pd.DataFrame,
           pd.DataFrame,
           pd.DataFrame,
           List[datatable.Column],
           List[datatable.Column]]:
    """Loads the data tables used to run ``funpack``.

    :arg fileinfo:  :class:`.FileInfo` object describing the input data files.
    :arg varfiles:  Path to one or more partial variable table files
    :arg dcfiles:   Path to one or more partial data coding table files
    :arg typefile:  Path to the type table file
    :arg procfile:  Path to the processing table file
    :arg catfile:   Path to the category table file

    All other arguments are passed throughh to the :func:`loadVariableTable`
    and :func:`loadProcessingTable` functions.

    :returns:      A tuple containing:

                    - The variable table
                    - The processing table
                    - The category table
                    - List of :class:`.Column` objects representing columns
                      which were in the data file(s), but not in the variable
                      table.
                    - List of :class:`.Column` objects representing columns
                      which are uncategorised.
    """

    vartable, unk, _ = loadVariableTable(fileinfo,
                                         varfiles,
                                         dcfiles,
                                         typefile,
                                         **kw)
    proctable        = loadProcessingTable(procfile, **kw)
    cattable         = loadCategoryTable(catfile)
    unc              = identifyUncategorisedVariables(fileinfo, cattable)

    # uncategorised should not contain unknown
    unc = [c for c in unc if c not in unk]

    # Add categorise for unknown/
    # uncategorised variables
    addImplicitCategories(cattable, unk, unc)

    return vartable, proctable, cattable, unk, unc


def loadVariableTable(
        fileinfo    : finfo.FileInfo,
        varfiles    : Sequence[str]              = None,
        dcfiles     : Sequence[str]              = None,
        typefile    : str                        = None,
        noBuiltins  : bool                       = False,
        naValues    : Dict[int, str]             = None,
        childValues : Dict[int, Tuple[str, str]] = None,
        recoding    : Dict[int, Tuple[str, str]] = None,
        clean       : Dict[int, str]             = None,
        typeClean   : Dict[util.CTYPES, str]     = None,
        globalClean : str                        = None,
        dropAbsent  : bool                       = True,
        **kwargs  # pylint: disable=unused-argument
) -> Tuple[pd.DataFrame,
           Sequence[datatable.Column],
           Sequence[datatable.Column]]:
    """Given variable table and datacoding table file names, builds and returns
    the variable table.

    :arg fileinfo:    :class:`.FileInfo` object describing the input data
                      files.

    :arg varfiles:    Path(s) to one or more variable files

    :arg dcfiles:     Path(s) to one or more data coding files

    :arg typefile:    Path to the type file

    :arg noBuiltins:  If provided, the built-in variable and datacoding base
                      tables are not loaded.

    :arg naValues:    Dictionary of ``{vid : values}`` mappings, specifying
                      values which should be replaced with NA. The values
                      are expected to be strings of comma-separated values.

    :arg childValues: Dictionary of ``{vid : (exprs, values)}`` mappings,
                      specifying parent value expressions, and corresponding
                      child values. The expressions and values
                      are expected to be strings of comma-separated values
                      of the same length.

    :arg recoding:    Dictionary of ``{vid : (rawlevel, newlevel)}``
                      mappings. The raw and enw levels are expected to be
                      strings of comma-separated values of the same length.

    :arg clean:       Dictionary of ``{vid : expr}`` mappings containing
                      cleaning functions to apply - this will override
                      any cleaning specified in the variable file, and
                      any cleaning specified in ``typeClean``. The expressions
                      are expected to be strings.

    :arg typeClean:   Dictionary of ``{type : expr}`` mappings containing
                      cleaning functions to apply to all variables of a
                      specific type - this will override any cleaning
                      specified in the type file. The expressions
                      are expected to be strings.

    :arg globalClean: Expression containing cleaning functions to
                      apply to every variable - this will be performed after
                      variable-specific cleaning in the variable table,
                      or specified via ``clean`` or ``typeClean``. The
                      expressions are expected to be strings.

    :arg dropAbsent:  If ``True`` (the default), remove all variables from the
                      variable table which are not present in the data
                      file(s).

    All other keyword arguments are ignored.

    :returns: A tuple containing:

                - A ``pandas.DataFrame`` containing the variable table

                - A sequence of :class:`.Column` objects representing variables
                  which were present in the data files, but not in the variable
                  table, but were added to the variable table.

                - A sequence of :class:`.Column` objects representing variables
                  which were present in the data files and in the variable
                  table, but which did not have any cleaning rules specified.
    """

    if noBuiltins: varbase, dcbase = None, None
    else:          varbase, dcbase = loadTableBases()

    vartable = mergeTableFiles(varbase,
                               varfiles,
                               'variable',
                               VARTABLE_DTYPES,
                               VARTABLE_CONVERTERS,
                               VARTABLE_COLUMNS)
    dctable  = mergeTableFiles(dcbase,
                               dcfiles,
                               'data coding',
                               DCTABLE_DTYPES,
                               {},
                               DCTABLE_COLUMNS)
    tytable  = mergeTableFiles(None,
                               [typefile],
                               'type',
                               TYPETABLE_DTYPES,
                               TYPETABLE_CONVERTERS,
                               TYPETABLE_COLUMNS)

    # Make sure data types are aligned,
    # otherwise we may run into problems
    # when merging them together.
    vartable = vartable.astype(
        {c : t for c, t in VARTABLE_DTYPES .items() if c != 'ID'})
    dctable  = dctable .astype(
        {c : t for c, t in DCTABLE_DTYPES  .items() if c != 'ID'})
    tytable  = tytable .astype(
        {c : t for c, t in TYPETABLE_DTYPES.items() if c != 'Type'})

    vartable.index = vartable.index.astype(VARTABLE_DTYPES[ 'ID'])
    dctable .index = dctable .index.astype(DCTABLE_DTYPES[  'ID'])
    tytable .index = tytable .index.astype(TYPETABLE_DTYPES['Type'])

    # Build a list of all columns in the input
    # data files, with the index column(s)
    # from each file dropped (index columns
    # are assigned a VID of 0)
    cols = []
    for df in fileinfo.datafiles:
        dfcols = fileinfo.columns(df)
        cols.extend([c for c in dfcols if c.vid != 0])

    # Make sure the variable table
    # contains an entry for every
    # variable in the input data.
    unknownVars = sanitiseVariableTable(vartable, cols, dropAbsent)

    # Merge data coding specific NAValues,
    # RawLevels, and NewLevels from the data
    # coding table into the variable table.
    mergeDataCodingTable(vartable, dctable)

    # Merge provided naValues, recodings,
    # and childValues into the variable
    # table (overriding whatever was specified
    # in the datacoding/variable tables)
    if naValues is not None:
        mergeIntoVariableTable(
            vartable,
            'NAValues',
            naValues)

    if recoding is not None:
        mergeIntoVariableTable(
            vartable,
            ['RawLevels', 'NewLevels'],
            recoding)

    if childValues is not None:
        childValues = {vid : (convert_ParentValues(expr), values)
                       for vid, (expr, values) in childValues.items()}
        mergeIntoVariableTable(
            vartable,
            ['ParentValues', 'ChildValues'],
            childValues)

    # navalues, raw/new levels and child values
    # are all still comma-separated strings -
    # convert them to types appropriate to the
    # datafield/variable
    def convert(rowvalues, column):
        val = rowvalues[column]

        if pd.isna(val):
            return np.nan

        val   = convert_comma_sep_text(val)
        ctype = rowvalues['Type']
        dtype = util.DATA_TYPES.get(ctype, None)

        if ctype in (util.CTYPES.date, util.CTYPES.time):
            return pd.to_datetime(val).values
        else:
            return np.array(val, dtype=dtype)

    # Make sure the series type stays as "object" - if
    # no rules are specified on any vid, the convert
    # function will return all nans, and pandas will
    # coerce the series type to float64, which may
    # interfere with subsequent rule assignments (as
    # these series may contain numpy arrays or
    # list-likes)
    if len(vartable) > 0:
        for col in ('NAValues', 'RawLevels', 'NewLevels', 'ChildValues'):
            newseries     = vartable.apply(convert, axis=1, args=(col,))
            vartable[col] = newseries.astype(object)

    # Before merging the cleaning functions
    # in, we generate a list of variables
    # which are "uncleaned", i.e. have not
    # had any cleaning specified, as this
    # may indicate that a variable has been
    # overlooked.
    #
    # If a variable has indirectly had NA
    # value insertion or recoding applied
    # via its data coding, it is not included
    # in this list.
    if clean is not None: ucmask = ~vartable.index.isin(clean.keys())
    else:                 ucmask =  vartable.index.notna()

    ucmask      = (ucmask                          &
                   vartable['NAValues']    .isna() &
                   vartable['RawLevels']   .isna() &
                   vartable['ParentValues'].isna() &
                   vartable['Clean']       .isna())
    ucmask      = ucmask[ucmask]
    uncleanVars = [c for c in cols
                   if (c.vid in ucmask.index and
                       c not in unknownVars)]

    # Merge clean options into variable table
    mergeCleanFunctions(vartable, tytable, clean, typeClean, globalClean)

    # Check where we can that the
    # vartable contains valid rules
    def checkLengths(col1, col2, row):
        val1  = row[col1]
        val2  = row[col2]
        isna1 = pd.isna(val1)
        isna2 = pd.isna(val2)

        # ugh. if the value is a sequence, isna
        # will return a sequence of bools
        if not isinstance(isna1, bool): isna1 = False
        if not isinstance(isna2, bool): isna2 = False

        if isna1 and isna2:
            return
        if isna1 or isna2 or (len(val1) != len(val2)):
            raise ValueError('Columns don\'t match [len({}) != '
                             'len({})]: {}'.format(val1, val2, row.name))

    checkRecoding     = ft.partial(checkLengths, 'RawLevels',    'NewLevels')
    checkParentValues = ft.partial(checkLengths, 'ParentValues', 'ChildValues')

    vartable.apply(checkRecoding,     axis=1)
    vartable.apply(checkParentValues, axis=1)

    return vartable, unknownVars, uncleanVars


def loadTableBases() -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Loads the UK Biobank variable and data coding schema files.

    This function is called by :func:`loadVariableTable`. It loads the UK
    Biobank variable and data coding schema files (available from the UK
    Biobank data showcase web site), and returns the information contained
    within as two ``pandas.Dataframe`` objects. These dataframes are then
    used as bases for the ``funpack`` variable table.

    Information in the base tables is loaded from the following files:

     - ``field.txt``:    A list of all UKB variables
     - ``encoding.txt``: A list of all UKB data codings
     - ``type.txt``:     A list of ``vid : type`` mappings for certain
                         variables, where ``type`` is the name of a
                         ``numpy`` data type (e.g. ``float32``).

    :returns: A tuple containing:

               - a ``pandas.DataFrame`` to be used as the base for the
                 variable table
               - a ``pandas.DataFrame`` to be used as the base for the
                 datacoding table
    """

    # Here we load these files, both obtained
    # from the UK Biobank showcase website:
    #
    #  - field.txt    - describes all UK biobank variables
    #  - encoding.txt - describes all data codings
    #
    # And we also load type.txt, which contains
    # the internal type to use for some variables

    # This dict contains all possible combinations
    # of (value_type, base_type) from field.txt
    typecodes = {
        (11,  0)  : util.CTYPES.integer,
        (31,  0)  : util.CTYPES.continuous,
        (21,  11) : util.CTYPES.categorical_single,
        (21,  41) : util.CTYPES.categorical_single_non_numeric,
        (22,  11) : util.CTYPES.categorical_multiple,
        (22,  41) : util.CTYPES.categorical_multiple_non_numeric,
        (61,  0)  : util.CTYPES.time,
        (51,  0)  : util.CTYPES.date,
        (41,  0)  : util.CTYPES.text,
        (101, 0)  : util.CTYPES.compound,
    }

    # We need pandas >=0.24 to support enums here
    def settype(valtype, basetype):
        return typecodes[valtype, basetype]

    datadir   = op.join(op.dirname(__file__), 'data')
    fields    = op.join(datadir, 'field.txt')
    encodings = op.join(datadir, 'encoding.txt')
    types     = op.join(datadir, 'type.txt')

    fields    = pd.read_csv(fields,    delimiter='\t')
    encodings = pd.read_csv(encodings, delimiter='\t')
    types     = pd.read_csv(types,     delimiter='\t',
                            index_col=0,
                            converters={'Type' : convert_dtype})

    dcbase  = pd.DataFrame({'ID' : encodings['encoding_id']}).set_index('ID')
    varbase = pd.DataFrame({
        'ID'           : fields['field_id'].astype(np.uint64),
        'Type'         : fields['value_type'].combine(fields['base_type'],
                                                      settype),
        'Description'  : fields['title'],
        'DataCoding'   : fields['encoding_id'],
        'Instancing'   : fields['instance_id'],
    }).set_index('ID')

    types.rename({'Type' : 'InternalType'}, axis=1, inplace=True)
    varbase = pd.concat((varbase, types), axis=1, join='outer')

    return varbase, dcbase


def mergeTableFiles(
        base       : pd.DataFrame,
        fnames     : List[str],
        what       : str,
        dtypes     : Dict[str, Type],
        converters : Dict[str, Callable],
        columns    : List[str]
) -> pd.DataFrame:
    """Load and merge one or more table files.

    This function is called by :func:`loadVariableTable` to load the variable,
    data coding, and type table files.

    The variable and datacoding tables can be loaded from multiple files, with
    each file containing part of the full table. All provided files are merged
    into one table. The final table for a given set of files is the outer join
    on the index column (assumed to be the first column in each file), where
    non-na values in overlapping columns from later files will overwrite the
    values in earlier files.

    :arg base:       Table containing base information - used for the variable
                     and datacoding tables (see :func:`_loadTableBases`).

    :arg fnames:     List of files to load. If ``None``, the ``base`` table
                     is returned (or an empty table if a ``base`` is not
                     given).

    :arg what:       Name of the table files being loaded - used solely for log
                     messages

    :arg dtypes:     Dict containing ``{column : datatype}`` mappings

    :arg converters: Dict containing ``{column : convertfunc}`` mappings

    :arg columns:    Expected column names
    """

    idcol   = columns[0]
    columns = columns[1:]

    if fnames is None:
        fnames = []

    fnames = [f for f in fnames if f is not None]

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=pd.errors.ParserWarning)

        for f in fnames:

            log.debug('Loading %s table from %s', what, f)

            table = pd.read_csv(f,
                                sep='\t',
                                index_col=0,
                                dtype=dtypes,
                                converters=converters)

            if base is None:
                base = table
                continue

            # Merge each file with an outer
            # join, so we retain all IDs and
            # columns defined across the
            # entire set of files.
            merged = base.merge(table,
                                how='outer',
                                on=idcol,
                                sort=True,
                                suffixes=('_x', '_y'))

            # Now we merge overlapping columns -
            # non-na values in later files take
            # precedence.
            for c in [c[:-2] for c in merged.columns if c.endswith('_x')]:
                bname                = c + '_x'
                tname                = c + '_y'
                merged[c]            = merged[bname]
                notna                = merged[tname].notna()
                merged.loc[notna, c] = merged.loc[notna, tname]
                merged               = merged.drop(columns=[bname, tname])
            base = merged

    # no base, and no files
    if base is None:
        base = pd.DataFrame()

    # error if we have any
    # unrecognised columns
    for col in base.columns:
        if col not in columns:
            raise ValueError('Unrecognised column in table file {} - '
                             'should be {}, but file contained {}.'.format(
                                 fnames, columns, base.columns))

    # in-fill any columns that
    # were not provided
    for col in columns:
        if col not in base.columns:
            base[col] = pd.Series(dtype=np.float64)

    return base


def sanitiseVariableTable(
        vartable   : pd.DataFrame,
        cols       : Sequence[datatable.Column],
        dropAbsent : bool
) -> List[datatable.Column]:
    """Ensures that the variable table contains an entry for every
    variable in the input data.

    Called by :func:`loadVariableTable`.

    :arg vartable:   ``pandas.DataFrame`` containing the variable table.

    :arg cols:       Sequence of :class:`.Column` objects representing
                     the columns in the input data.

    :arg dropAbsent: If ``True``, entries in the table for variables which are
                     not in ``cols`` will be removed.

    :return:         A list of unknown :class:`.Column` objects, i.e.
                     representing variables which were not in the variable
                     table.
    """

    unknownVars = []

    # Make sure a placeholder entry is
    # present for any variables which are
    # not in the variable table, but which
    # are in the data file(s).
    for col in cols:

        vid  = col.vid
        name = col.name

        if vid in vartable.index:
            continue

        unknownVars.append(col)
        addNewVariable(vartable, vid, name)

    # And the inverse - we can drop any
    # variables from the variable table
    # that are not in the data.
    if dropAbsent:
        vids = [c.vid for c in cols]
        vartable.drop([v for v in vartable.index if v not in vids],
                      inplace=True)
    return unknownVars


def mergeIntoVariableTable(
        vartable : pd.DataFrame,
        cols     : Sequence[str],
        mapping  : Union[str, Dict[int, Any]]):
    """Merge data from ``mapping`` into the variable table.

    Called by :func:`loadVariableTable`.

    :arg vartable: The variable table

    :arg cols:     Names of columns in the variable table

    :arg mapping:  Dict of ``{vid : values}`` mappings containing the
                   data to copy in.
    """

    onecol = isinstance(cols, str)
    if onecol:
        cols = [cols]

    # Ignore any variables that
    # are not in variable table
    vids   = list(mapping.keys())
    vin    = pd.Series(vids).isin(vartable.index)
    vids   = [v for i, v in enumerate(vids) if vin[i]]

    for vid in vids:
        vals = mapping[vid]

        if onecol:
            vals = [vals]

        for col, val in zip(cols, vals):
            vartable.at[vid, col] = val


def mergeDataCodingTable(
        vartable : pd.DataFrame,
        dctable  : pd.DataFrame):
    """Merges information from the data coding table into the variable
    table.

    Called by :func:`loadVariableTable`.

    :arg vartable: The variable table.
    :arg dctable:  The data coding table.
    """

    with_datacoding = vartable['DataCoding'].notna()

    for field in ['NAValues', 'RawLevels', 'NewLevels']:
        mask    = vartable[field].isna() & with_datacoding
        newvals = vartable.loc[mask].merge(dctable,
                                           left_on='DataCoding',
                                           right_index=True,
                                           suffixes=('_v', '_dc'),
                                           copy=False)['{}_dc'.format(field)]
        vartable.loc[mask, field] = newvals


def mergeCleanFunctions(
        vartable    : pd.DataFrame,
        tytable     : pd.DataFrame,
        clean       : Dict[int, str],
        typeClean   : Dict[str, str],
        globalClean : str):
    """Merges custom clean functions into the variable table.

    Called by :func:`loadVariableTable`.

    :arg vartable:    The variable table.

    :arg tytable:     The type table

    :arg clean:       Dictionary of ``{vid : expr}`` mappings containing
                      cleaning functions to apply - this will override
                      any cleaning specified in the variable file, and
                      any cleaning specified in ``typeClean``.

    :arg typeClean:   Dictionary of ``{type : expr}`` mappings containing
                      cleaning functions to apply to all variables of a
                      specific type - this will override any cleaning
                      specified in the type file.

    :arg globalClean: Expression containing cleaning functions to
                      apply to every variable - this will be performed after
                      variable-specific cleaning in the variable table,
                      or specified via ``clean`` or ``typeClean``.
    """

    # Merge type-specific Clean
    # from the type table into
    # the variable table.
    for vid in vartable.index:

        if vid == 0:
            continue

        vtype = vartable.loc[vid, 'Type']
        pp    = vartable.loc[vid, 'Clean']

        # Override with typeClean if necessary
        if typeClean is not None and vtype in typeClean:
            tpp = convert_Process('cleaner', typeClean[vtype])
        elif vtype in tytable.index:
            tpp = collections.OrderedDict((tytable.loc[vtype, 'Clean']))
        else:
            continue

        # type cleaning is applied after
        # variable-specific cleaning
        if pd.isnull(pp): vartable.loc[[vid], 'Clean'] = [tpp]
        else:             vartable.loc[ vid,  'Clean'].update(tpp)

    # Override cleaning with expressions
    # that have been passed on the command line
    if clean is not None:
        clean = {vid : convert_Process('cleaner', expr)
                 for vid, expr in clean.items()}
        mergeIntoVariableTable(vartable, 'Clean', clean)

    # Add global cleaning to all variables
    if globalClean is not None:

        for vid in vartable.index:

            if vid == 0:
                continue

            pp  = vartable.loc[vid, 'Clean']
            gpp = convert_Process('cleaner', globalClean)

            # global cleaning is applied
            # after all other cleaning
            if pd.isnull(pp): vartable.loc[[vid], 'Clean'] = [gpp]
            else:             vartable.loc[ vid,  'Clean'].update(gpp)


def addNewVariable(
        vartable   : pd.DataFrame,
        vid        : int,
        name       : str,
        dtype      : np.dtype = None,
        instancing : int      = None):
    """Add a new row to the variable table.

    The ``instancing`` argument defines the meaning of the
    :attr:`.Column.visit` field for columns associated with this variable.
    The default value is ``2``, meaning that this variable may be associated
    with columns corresponding to measurements acquired at different UK
    BioBank assessments and imaging visits.  See
    https://biobank.ctsu.ox.ac.uk/crystal/schema.cgi?id=9 and
    https://biobank.ctsu.ox.ac.uk/crystal/schema.cgi?id=10 for more details.

    .. note:: If an entry for the specified ``vid`` already exists in
              ``vartable``, the ``name``, ``dtype`` and ``instancing``
              arguments are ignored and the existing information in
              ``vartable`` will take precedence.

    :arg vartable:   The variable table

    :arg vid:        Integer variable ID

    :arg name:       Variable name - used as the description

    :arg dtype:      ``numpy`` data type. If ``None``, the variable type
                     is set to :attr:`.util.CTYPES.unknown`.

    :arg instancing: Instancing code for the new variable - defaults to ``2``
    """

    # If an entry already exists
    # in the variable table, it
    # takes precedence
    if vid in vartable.index:
        dtype      = vartable.at[vid, 'Type']
        name       = vartable.at[vid, 'Description']
        instancing = vartable.at[vid, 'Instancing']

    else:
        # set dtype to something which
        # will cause the conditionals
        # to fall through
        if dtype is None: dtype = object
        else:             dtype = dtype.type

        # Assume that new variables
        # are associated with visits
        if instancing is None:
            instancing = 2

        if   issubclass(dtype, np.integer): dtype = util.CTYPES.integer
        elif issubclass(dtype, np.float):   dtype = util.CTYPES.continuous
        else:                               dtype = util.CTYPES.unknown

    vartable.loc[vid, 'Description'] = name
    vartable.loc[vid, 'Type']        = dtype
    vartable.loc[vid, 'Instancing']  = instancing


def loadProcessingTable(
        procfile       : str                             = None,
        skipProcessing : bool                            = False,
        prependProcess : Sequence[Tuple[List[int], str]] = None,
        appendProcess  : Sequence[Tuple[List[int], str]] = None,
        **kwargs  # pylint: disable=unused-argument
) -> pd.DataFrame:
    """Loads the processing table from the given file.

    :arg procfile:       Path to the processing table file.

    :arg skipProcessing: If ``True``, the processing table is not loaded from
                         ``procfile``. The ``prependProcess`` and
                         ``appendProcess`` arguments are still applied.

    :arg prependProcess: Sequence of ``(varids, procstr)`` mappings specifying
                         processes to prepend to the beginning of the
                         processing table.

    :arg appendProcess:  Sequence of ``(varids, procstr)`` mappings specifying
                         processes to append to the end of the processing
                         table.

    All other keyword arguments are ignored.
    """

    if prependProcess is None: prependProcess = []
    if appendProcess  is None: appendProcess  = []

    if (procfile is not None) and (not skipProcessing):
        log.debug('Loading processing table from %s', procfile)
        proctable = pd.read_csv(procfile,
                                sep='\t',
                                index_col=False,
                                skip_blank_lines=True,
                                comment='#',
                                converters=PROCTABLE_CONVERTERS)

    else:
        proctable = pd.DataFrame(columns=PROCTABLE_COLUMNS)

    # prepend/append custom
    # processes to the table
    proctable.index += len(prependProcess)
    for i, (vids, procs) in enumerate(prependProcess):
        vids  = convert_Process_Variable(vids)
        procs = convert_Process('processor', procs)
        proctable.at[i, 'Variable'] = np.array(vids, dtype=object)
        proctable.at[i, 'Process']  = procs

    for i, (vids, procs) in enumerate(appendProcess, len(proctable.index)):
        vids  = convert_Process_Variable(vids)
        procs = convert_Process('processor', procs)
        proctable.at[i, 'Variable'] = np.array(vids, dtype=object)
        proctable.at[i, 'Process']  = procs

    proctable.sort_index(inplace=True)

    return proctable


def loadCategoryTable(catfile : str = None) -> pd.DataFrame:
    """Loads the category table from the given file.

    :arg catfile: Path to the category file.
    """
    if catfile is not None:
        log.debug('Loading category table from %s', catfile)
        cattable = pd.read_csv(catfile,
                               sep='\t',
                               index_col=0,
                               dtype=CATTABLE_DTYPES,
                               converters=CATTABLE_CONVERTERS)
    else:
        cattable            = pd.DataFrame(columns=CATTABLE_COLUMNS[1:])
        cattable.index.name = CATTABLE_COLUMNS[0]


    return cattable


def categoryVariables(
        cattable   : pd.DataFrame,
        categories : Sequence[Union[int, str]]
) -> List[int]:
    """Returns a list of variable IDs from ``cattable`` which correspond to
    the strings in ``categories``.

    :arg cattable:   The category table.
    :arg categories: Sequence of integer category IDs or label sub-strings
                     specifying the categories to return.
    :returns:        A list of variable IDs.
    """

    allvars = []

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        for cat in categories:

            catpat  = re.compile('({})'.format(cat), re.IGNORECASE)
            idmask  = cattable.index.isin([cat])
            lblmask = cattable['Category'].str.contains(catpat)
            catvars = cattable.loc[idmask | lblmask, 'Variables']

            if len(catvars) == 0:
                continue

            for c in catvars.iloc[0]:
                if c not in allvars:
                    allvars.append(c)

    return allvars


def variableCategories(
        cattable : pd.DataFrame,
        vids     : Sequence[int]
) -> Dict[int, List[str]]:
    """Identify the categories for each variable in vids.

    :arg cattable: The category table
    :arg vids:     Sequence of variable/datafield IDs
    :returns:      A dict of ``{vid : [category]}`` mappings
    """

    categories = collections.defaultdict(list)

    for v in vids:
        for i in cattable.index:
            if v in cattable.loc[i, 'Variables']:
                categories[v].append(cattable.loc[i, 'Category'])

    return categories


def addImplicitCategories(
        cattable : pd.DataFrame,
        unknown  : Sequence[datatable.Column],
        uncat    : Sequence[datatable.Column]):
    """Adds some implicit/automatic categories to the category table.

    The following implicit categories are added:

     - ``unknown``:       Variables which are not present in the variable
                          table - this comprises non-UKB variables, or new UKB
                          variables which are not described by the internal
                          FUNPACK variable information in ``funpack/data/``.

     - ``uncategorised``: Variables which are not present in any other
                          category.

    :arg cattable: The category table.
    :arg unknown:  Sequence of :class:`.Column` objects representing
                   variables to add to an "unknown" category.
    :arg uncat:    Sequence of :class:`.Column` objects representing
                   variables to add to an "uncategorised" category.
    """

    for cols, label in zip((unknown, uncat), ('unknown', 'uncategorised')):

        if cols is None:
            continue

        vids = list(sorted({c.vid for c in cols}))

        # label already in table?
        umask = cattable['Category'] == label

        if np.any(umask):
            idx  = np.where(umask)[0][0]
            idx  = cattable.index[idx]
            vids = cattable.loc[idx, 'Variables'] + vids
        else:
            idx = IMPLICIT_CATEGORIES[label]

        cattable.loc[idx, 'Category']  = label
        cattable.loc[idx, 'Variables'] = list(vids)


def columnTypes(
        vartable : pd.DataFrame,
        columns  : Sequence[datatable.Column]
) -> Tuple[List[util.CTYPES], Dict[str, np.dtype]]:
    """Retrieves the type of each column in ``cols`` as listed in ``vartable``.
    Also identifies a suitable internal data type to use for each column where
    possible.

    :arg vartable: The variable table.

    :arg columnss: List of :class:`.Column` objects.

    :returns:      A tuple containing:

                    - A list containing the type for each column in ``cols`` -
                      an identifier from the :attr:`.util.CTYPES` enum.
                      Columns corresponding to a variable which is not in
                      the variable table is given a type of ``None``.

                    - A dict of ``{ column_name : dtype }`` mappings containing
                      a suitable internal data type to use for some columns.
    """

    vttypes = []
    dtypes  = {}

    for col in columns:

        vid  = col.vid
        name = col.name

        if vid not in vartable.index:
            vttypes.append(None)
            continue

        vttype = vartable.loc[vid, 'Type']
        dtype  = vartable.loc[vid, 'InternalType']

        if pd.isna([dtype])[0]:
            dtype = util.DATA_TYPES.get(vttype, None)

        vttypes.append(vttype)
        if dtype is not None:
            dtypes[name] = dtype

    return vttypes, dtypes


def identifyUncategorisedVariables(
        fileinfo : finfo.FileInfo,
        cattable : pd.DataFrame
) -> List[datatable.Column]:
    """Called by :func:`loadTables`. Identifies all variables which are in the
    data file(s), but which are uncategorised (not present in any categories
    in the category table).

    Such variables might have been overlooked, so the user may need to be
    warned about them.

    :arg fileinfo:  :class:`.FileInfo` object.
    :arg cattable:  Category table
    :returns:       A list of :class:`.Column` objects associated with
                    variables that are uncategorised.
    """

    def isCategorised(col):
        def inCategory(catvars):
            return col.vid in catvars
        return cattable['Variables'].apply(inCategory).any()

    uncategorised = []
    for datafile in fileinfo.datafiles:
        cols = fileinfo.columns(datafile)
        cols = [c for c in cols if (c.vid != 0) and (not isCategorised(c))]
        uncategorised.extend(cols)

    return uncategorised
